import autolens as al
import os
import shutil

import numpy as np
import pytest
from autolens import exc

test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestNewUVPlaneDataFrom(object):
    def test__new_data_with_resized_primary_beam(self):

        uv_plane_data = al.UVPlaneData(
            shape=(2, 2),
            visibilities=np.array([[1, 1]]),
            pixel_scale=1.0,
            primary_beam=al.PrimaryBeam(np.zeros((5, 5)), pixel_scale=1.0),
            noise_map=1,
            exposure_time_map=1,
            uv_wavelengths=1,
        )

        uv_plane_data = uv_plane_data.new_uv_plane_data_with_resized_primary_beam(
            new_shape=(1, 1)
        )
        assert uv_plane_data.pixel_scale == 1.0
        assert (uv_plane_data.primary_beam == np.zeros((1, 1))).all()
        assert uv_plane_data.origin == (0.0, 0.0)

    def test__new_data_with_modified_visibilities(self):

        uv_plane_data = al.UVPlaneData(
            shape=(2, 2),
            visibilities=np.array([[1, 1]]),
            pixel_scale=1.0,
            primary_beam=al.PrimaryBeam(np.zeros((5, 5)), pixel_scale=1.0),
            noise_map=1,
            exposure_time_map=2,
            uv_wavelengths=3,
        )

        uv_plane_data = uv_plane_data.new_uv_plane_data_with_modified_visibilities(
            modified_visibilities=np.array([[2, 2]])
        )

        assert (uv_plane_data.visibilities == np.array([[2, 2]])).all()
        assert uv_plane_data.shape == (2, 2)
        assert uv_plane_data.pixel_scale == 1.0
        assert (uv_plane_data.primary_beam == np.zeros((1, 1))).all()
        assert uv_plane_data.noise_map == 1
        assert uv_plane_data.exposure_time_map == 2
        assert uv_plane_data.uv_wavelengths == 3
        assert uv_plane_data.origin == (0.0, 0.0)

    def test__new_data_in_counts__all_arrays_in_units_of_flux_are_converted(self):

        uv_plane_data = al.UVPlaneData(
            shape=(2, 2),
            visibilities=np.ones((3, 2)),
            pixel_scale=1.0,
            noise_map=2.0 * np.ones((3,)),
            exposure_time_map=0.5 * np.ones((3,)),
            primary_beam=1,
            uv_wavelengths=1,
        )

        uv_plane_data = uv_plane_data.new_uv_plane_data_converted_from_electrons()

        assert (uv_plane_data.visibilities == 2.0 * np.ones((3, 2))).all()
        assert (uv_plane_data.noise_map == 4.0 * np.ones((3,))).all()
        assert uv_plane_data.origin == (0.0, 0.0)

    def test__new_data_in_adus__all_arrays_in_units_of_flux_are_converted(self):

        uv_plane_data = al.UVPlaneData(
            shape=(2, 2),
            visibilities=np.ones((3, 2)),
            pixel_scale=1.0,
            noise_map=2.0 * np.ones((3,)),
            exposure_time_map=0.5 * np.ones((3)),
            primary_beam=1,
            uv_wavelengths=1,
        )

        uv_plane_data = uv_plane_data.new_uv_plane_data_converted_from_adus(gain=2.0)

        assert (uv_plane_data.visibilities == 2.0 * 2.0 * np.ones((3, 2))).all()
        assert (uv_plane_data.noise_map == 2.0 * 4.0 * np.ones((3,))).all()
        assert uv_plane_data.origin == (0.0, 0.0)


class TestPrimaryBeam(object):
    class TestConstructors(object):
        def test__init__input_primary_beam__all_attributes_correct_including_data_inheritance(
            self
        ):
            psf = al.PrimaryBeam(
                array=np.ones((3, 3)), pixel_scale=1.0, renormalize=False
            )

            assert psf.shape == (3, 3)
            assert psf.pixel_scale == 1.0
            assert (psf == np.ones((3, 3))).all()
            assert psf.origin == (0.0, 0.0)

            psf = al.PrimaryBeam(
                array=np.ones((4, 3)), pixel_scale=1.0, renormalize=False
            )

            assert (psf == np.ones((4, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.shape == (4, 3)
            assert psf.origin == (0.0, 0.0)

        def test__from_fits__input_primary_beam_3x3__all_attributes_correct_including_data_inheritance(
            self
        ):
            psf = al.PrimaryBeam.from_fits_with_scale(
                file_path=test_data_dir + "3x3_ones.fits", hdu=0, pixel_scale=1.0
            )

            assert (psf == np.ones((3, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.origin == (0.0, 0.0)

            psf = al.PrimaryBeam.from_fits_with_scale(
                file_path=test_data_dir + "4x3_ones.fits", hdu=0, pixel_scale=1.0
            )

            assert (psf == np.ones((4, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.origin == (0.0, 0.0)

    class TestRenormalize(object):
        def test__input_is_already_normalized__no_change(self):
            primary_beam_data = np.ones((3, 3)) / 9.0

            psf = al.PrimaryBeam(
                array=primary_beam_data, pixel_scale=1.0, renormalize=True
            )

            assert psf == pytest.approx(primary_beam_data, 1e-3)

        def test__input_is_above_normalization_so_is_normalized(self):

            primary_beam_data = np.ones((3, 3))

            psf = al.PrimaryBeam(
                array=primary_beam_data, pixel_scale=1.0, renormalize=True
            )

            assert psf == pytest.approx(np.ones((3, 3)) / 9.0, 1e-3)

        def test__same_as_above__renomalized_false_does_not_renormalize(self):
            primary_beam_data = np.ones((3, 3))

            psf = al.PrimaryBeam(
                array=primary_beam_data, pixel_scale=1.0, renormalize=False
            )

            assert psf == pytest.approx(np.ones((3, 3)), 1e-3)

    class TestBinnedUp(object):
        def test__primary_beam_is_even_x_even__rescaled_to_odd_x_odd__no_use_of_dimension_trimming(
            self
        ):

            array_2d = np.ones((6, 6))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((9, 9))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.333333333333333, renormalize=True
            )
            assert psf.pixel_scale == 3.0
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((18, 6))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 27.0) * np.ones((9, 3))

            array_2d = np.ones((6, 18))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 27.0) * np.ones((3, 9))

        def test__primary_beam_is_even_x_even_after_binning_up__resized_to_odd_x_odd_with_shape_plus_one(
            self
        ):

            array_2d = np.ones((2, 2))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == 0.4
            assert psf == (1.0 / 25.0) * np.ones((5, 5))

            array_2d = np.ones((40, 40))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.1, renormalize=True
            )
            assert psf.pixel_scale == 8.0
            assert psf == (1.0 / 25.0) * np.ones((5, 5))

            array_2d = np.ones((2, 4))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(0.4444444, 1.0e-4)
            assert psf == (1.0 / 45.0) * np.ones((5, 9))

            array_2d = np.ones((4, 2))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(0.4444444, 1.0e-4)
            assert psf == (1.0 / 45.0) * np.ones((9, 5))

        def test__primary_beam_is_odd_and_even_after_binning_up__resized_to_odd_and_odd_with_shape_plus_one(
            self
        ):

            array_2d = np.ones((6, 4))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(2.0, 1.0e-4)
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((9, 12))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.33333333333, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(3.0, 1.0e-4)
            assert psf == (1.0 / 15.0) * np.ones((3, 5))

            array_2d = np.ones((4, 6))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(2.0, 1.0e-4)
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((12, 9))
            psf = al.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.33333333333, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(3.0, 1.0e-4)
            assert psf == (1.0 / 15.0) * np.ones((5, 3))

    class TestNewRenormalizedPrimaryBeam(object):
        def test__input_is_already_normalized__no_change(self):

            primary_beam_data = np.ones((3, 3)) / 9.0

            psf = al.PrimaryBeam(
                array=primary_beam_data, pixel_scale=1.0, renormalize=False
            )

            primary_beam_new = psf.new_primary_beam_with_renormalized_array()

            assert primary_beam_new == pytest.approx(primary_beam_data, 1e-3)

        def test__input_is_above_normalization_so_is_normalized(self):

            primary_beam_data = np.ones((3, 3))

            psf = al.PrimaryBeam(
                array=primary_beam_data, pixel_scale=1.0, renormalize=False
            )

            primary_beam_new = psf.new_primary_beam_with_renormalized_array()

            assert primary_beam_new == pytest.approx(np.ones((3, 3)) / 9.0, 1e-3)

    class TestConvolve(object):
        def test__kernel_is_not_odd_x_odd__raises_error(self):
            kernel = np.array([[0.0, 1.0], [1.0, 2.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            with pytest.raises(exc.ConvolutionException):
                psf.convolve(np.ones((5, 5)))

        def test__image_is_3x3_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (blurred_image == kernel).all()

        def test__image_is_4x4_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [0.0, 1.0, 0.0, 0.0],
                        [1.0, 2.0, 1.0, 0.0],
                        [0.0, 1.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0],
                    ]
                )
            ).all()

        def test__image_is_4x3_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]]
                )
            ).all()

        def test__image_is_3x4_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [[0.0, 1.0, 0.0, 0.0], [1.0, 2.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0]]
                )
            ).all()

        def test__image_is_4x4_has_two_central_values__kernel_is_asymmetric__blurred_image_follows_convolution(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [1.0, 1.0, 1.0, 0.0],
                        [2.0, 3.0, 2.0, 1.0],
                        [1.0, 5.0, 5.0, 1.0],
                        [0.0, 1.0, 3.0, 3.0],
                    ]
                )
            ).all()

        def test__image_is_4x4_values_are_on_edge__kernel_is_asymmetric__blurring_does_not_account_for_edge_effects(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [1.0, 1.0, 0.0, 0.0],
                        [2.0, 1.0, 1.0, 1.0],
                        [3.0, 3.0, 2.0, 2.0],
                        [0.0, 0.0, 1.0, 3.0],
                    ]
                )
            ).all()

        def test__image_is_4x4_values_are_on_corner__kernel_is_asymmetric__blurring_does_not_account_for_edge_effects(
            self
        ):
            image = np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = al.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [2.0, 1.0, 0.0, 0.0],
                        [3.0, 3.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 1.0],
                        [0.0, 0.0, 2.0, 2.0],
                    ]
                )
            ).all()

    class TestFromGaussian(object):
        def test__identical_to_gaussian_light_profile(self):

            grid = al.Grid.from_shape_pixel_scale_and_sub_size(
                shape=(3, 3), pixel_scale=1.0, sub_size=1
            )

            gaussian = al.light_profiles.EllipticalGaussian(
                centre=(0.1, 0.1), axis_ratio=0.9, phi=45.0, intensity=1.0, sigma=1.0
            )
            profile_gaussian = gaussian.profile_image_from_grid(
                grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
            )

            profile_psf = al.PrimaryBeam(
                array=profile_gaussian, pixel_scale=1.0, renormalize=True
            )

            imaging_psf = al.PrimaryBeam.from_gaussian(
                shape=(3, 3),
                pixel_scale=1.0,
                centre=(0.1, 0.1),
                axis_ratio=0.9,
                phi=45.0,
                sigma=1.0,
            )

            assert profile_psf == pytest.approx(imaging_psf, 1e-4)


class TestSimulateUVPlaneData(object):
    def test__setup_with_all_features_off(self, transformer_7x7_7):
        image = np.array([[2.0, 0.0, 0.0], [0.0, 1.0, 0.0], [3.0, 0.0, 0.0]])

        exposure_time_map = al.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        uv_plane_data_simulated = al.SimulatedUVPlaneData.from_image_and_exposure_arrays(
            image=image,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            pixel_scale=0.1,
            transformer=transformer_7x7_7,
            noise_sigma=None,
        )

        image_1d = np.array([2.0, 0.0, 0.0, 0.0, 1.0, 0.0, 3.0, 0.0, 0.0])
        simulated_visibilities = transformer_7x7_7.visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert uv_plane_data_simulated.visibilities == pytest.approx(
            simulated_visibilities, 1.0e-4
        )
        assert uv_plane_data_simulated.pixel_scale == 0.1

    def test__setup_with_background_sky_on__noise_off__no_noise_in_image__noise_map_is_noise_value(
        self, transformer_7x7_7
    ):
        image = np.array([[2.0, 0.0, 0.0], [0.0, 1.0, 0.0], [3.0, 0.0, 0.0]])

        exposure_time_map = al.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        background_sky_map = al.ScaledSquarePixelArray.single_value(
            value=2.0, pixel_scale=0.1, shape=image.shape
        )

        uv_plane_data_simulated = al.SimulatedUVPlaneData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            transformer=transformer_7x7_7,
            noise_sigma=None,
            noise_if_add_noise_false=0.2,
            noise_seed=1,
        )

        image_1d = np.array([4.0, 2.0, 2.0, 2.0, 3.0, 2.0, 5.0, 2.0, 2.0])
        simulated_visibilities = transformer_7x7_7.visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert uv_plane_data_simulated.visibilities == pytest.approx(
            simulated_visibilities, 1.0e-4
        )
        assert (
            uv_plane_data_simulated.exposure_time_map == 1.0 * np.ones((3, 3))
        ).all()
        assert uv_plane_data_simulated.pixel_scale == 0.1

        assert uv_plane_data_simulated.noise_map == 0.2 * np.ones((7, 2))

    def test__setup_with_noise(self, transformer_7x7_7):
        image = np.array([[2.0, 0.0, 0.0], [0.0, 1.0, 0.0], [3.0, 0.0, 0.0]])

        exposure_time_map = al.ScaledSquarePixelArray.single_value(
            value=20.0, pixel_scale=0.1, shape=image.shape
        )

        uv_plane_data_simulated = al.SimulatedUVPlaneData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=20.0,
            exposure_time_map=exposure_time_map,
            transformer=transformer_7x7_7,
            noise_sigma=0.1,
            noise_seed=1,
        )

        image_1d = np.array([2.0, 0.0, 0.0, 0.0, 1.0, 0.0, 3.0, 0.0, 0.0])
        simulated_visibilities = transformer_7x7_7.visibilities_from_image_1d(
            image_1d=image_1d
        )

        assert (
            uv_plane_data_simulated.exposure_time_map == 20.0 * np.ones((3, 3))
        ).all()
        assert uv_plane_data_simulated.pixel_scale == 0.1

        assert uv_plane_data_simulated.visibilities[0, :] == pytest.approx(
            [1.728611, -2.582958], 1.0e-4
        )
        visibilities_noise_map_realization = (
            uv_plane_data_simulated.visibilities - simulated_visibilities
        )

        assert visibilities_noise_map_realization == pytest.approx(
            uv_plane_data_simulated.noise_map_realization, 1.0e-4
        )

        assert uv_plane_data_simulated.noise_map == 0.1 * np.ones((7, 2))

    def test__from_deflections_and_galaxies__same_as_manual_calculation_using_tracer(
        self, transformer_7x7_7
    ):

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(10, 10), pixel_scale=1.0, sub_size=1
        )

        g0 = al.Galaxy(
            redshift=0.5,
            mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
        )

        g1 = al.Galaxy(
            redshift=1.0, light=al.light_profiles.SphericalSersic(intensity=1.0)
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

        deflections = tracer.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
        )

        uv_plane_data_simulated_via_deflections = al.SimulatedUVPlaneData.from_deflections_galaxies_and_exposure_arrays(
            deflections=deflections,
            pixel_scale=1.0,
            galaxies=[g1],
            exposure_time=10000.0,
            background_sky_level=100.0,
            transformer=transformer_7x7_7,
            noise_sigma=0.1,
            noise_seed=1,
        )

        tracer_profile_image = tracer.profile_image_from_grid(
            grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
        )

        uv_plane_data_simulated = al.SimulatedUVPlaneData.from_image_and_exposure_arrays(
            image=tracer_profile_image,
            pixel_scale=1.0,
            exposure_time=10000.0,
            background_sky_level=100.0,
            transformer=transformer_7x7_7,
            noise_sigma=0.1,
            noise_seed=1,
        )

        assert (
            uv_plane_data_simulated_via_deflections.exposure_time_map
            == uv_plane_data_simulated.exposure_time_map
        ).all()
        assert (
            uv_plane_data_simulated_via_deflections.visibilities
            == uv_plane_data_simulated.visibilities
        ).all()

        assert (
            uv_plane_data_simulated_via_deflections.noise_map
            == uv_plane_data_simulated.noise_map
        ).all()

    def test__from_tracer__same_as_manual_tracer_input(self, transformer_7x7_7):

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(20, 20), pixel_scale=0.05, sub_size=1
        )

        lens_galaxy = al.Galaxy(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic(intensity=1.0),
            mass=al.mass_profiles.EllipticalIsothermal(einstein_radius=1.6),
        )

        source_galaxy = al.Galaxy(
            redshift=1.0, light=al.light_profiles.EllipticalSersic(intensity=0.3)
        )

        tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

        uv_plane_data_simulated_via_tracer = al.SimulatedUVPlaneData.from_tracer_grid_and_exposure_arrays(
            tracer=tracer,
            grid=grid,
            pixel_scale=0.1,
            exposure_time=10000.0,
            background_sky_level=100.0,
            transformer=transformer_7x7_7,
            noise_sigma=0.1,
            noise_seed=1,
        )

        uv_plane_data_simulated = al.SimulatedUVPlaneData.from_image_and_exposure_arrays(
            image=tracer.profile_image_from_grid(
                grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
            ),
            pixel_scale=0.1,
            exposure_time=10000.0,
            background_sky_level=100.0,
            transformer=transformer_7x7_7,
            noise_sigma=0.1,
            noise_seed=1,
        )

        assert (
            uv_plane_data_simulated_via_tracer.exposure_time_map
            == uv_plane_data_simulated.exposure_time_map
        ).all()
        assert (
            uv_plane_data_simulated_via_tracer.visibilities
            == uv_plane_data_simulated.visibilities
        ).all()

        assert (
            uv_plane_data_simulated_via_tracer.noise_map
            == uv_plane_data_simulated.noise_map
        ).all()

    class TestCreateGaussianNoiseMap(object):
        def test__gaussian_noise_sigma_0__gaussian_noise_map_all_0__image_is_identical_to_input(
            self
        ):
            simulate_gaussian_noise = al.gaussian_noise_map_from_shape_and_sigma(
                shape=(9,), sigma=0.0, noise_seed=1
            )

            assert (simulate_gaussian_noise == np.zeros((9,))).all()

        def test__gaussian_noise_sigma_1__gaussian_noise_map_all_non_0__image_has_noise_added(
            self
        ):
            simulate_gaussian_noise = al.gaussian_noise_map_from_shape_and_sigma(
                shape=(9,), sigma=1.0, noise_seed=1
            )

            # Use seed to give us a known gaussian noises map we'll test for

            assert simulate_gaussian_noise == pytest.approx(
                np.array([1.62, -0.61, -0.53, -1.07, 0.87, -2.30, 1.74, -0.76, 0.32]),
                1e-2,
            )


class TestUVPlaneFromFits(object):
    def test__no_settings_just_pass_fits(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            noise_map_path=test_data_dir + "3_threes.fits",
            u_wavelengths_path=test_data_dir + "3_fours.fits",
            v_wavelengths_path=test_data_dir + "3_fives.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            renormalize_primary_beam=False,
        )

        assert (uv_plane_data.visibilities[:, 0] == np.ones(3)).all()
        assert (uv_plane_data.visibilities[:, 1] == 2.0 * np.ones(3)).all()
        assert (uv_plane_data.noise_map == 3.0 * np.ones(3)).all()
        assert (uv_plane_data.uv_wavelengths[:, 0] == 4.0 * np.ones(3)).all()
        assert (uv_plane_data.uv_wavelengths[:, 1] == 5.0 * np.ones(3)).all()
        assert (uv_plane_data.primary_beam == 5.0 * np.ones((3, 3))).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__optional_array_paths_included__loads_optional_array(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            renormalize_primary_beam=False,
        )

        assert (uv_plane_data.noise_map == 3.0 * np.ones((3,))).all()
        assert (uv_plane_data.primary_beam == 5.0 * np.ones((3, 3))).all()
        assert (uv_plane_data.exposure_time_map == 6.0 * np.ones((3,))).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__all_files_in_one_fits__load_using_different_hdus(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_multiple_hdu.fits",
            noise_map_hdu=2,
            primary_beam_path=test_data_dir + "3x3_multiple_hdu.fits",
            primary_beam_hdu=3,
            exposure_time_map_path=test_data_dir + "3x3_multiple_hdu.fits",
            exposure_time_map_hdu=5,
            real_visibilities_path=test_data_dir + "3_multiple_hdu.fits",
            real_visibilities_hdu=0,
            imaginary_visibilities_path=test_data_dir + "3_multiple_hdu.fits",
            imaginary_visibilities_hdu=1,
            u_wavelengths_path=test_data_dir + "3_multiple_hdu.fits",
            u_wavelengths_hdu=3,
            v_wavelengths_path=test_data_dir + "3_multiple_hdu.fits",
            v_wavelengths_hdu=4,
            renormalize_primary_beam=False,
        )

        assert (uv_plane_data.primary_beam == 4.0 * np.ones((3, 3))).all()
        assert (uv_plane_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()
        assert (uv_plane_data.visibilities[:, 0] == np.ones(3)).all()
        assert (uv_plane_data.visibilities[:, 1] == 2.0 * np.ones(3)).all()
        assert (uv_plane_data.noise_map == 3.0 * np.ones(3)).all()
        assert (uv_plane_data.uv_wavelengths[:, 0] == 4.0 * np.ones(3)).all()
        assert (uv_plane_data.uv_wavelengths[:, 1] == 5.0 * np.ones(3)).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__exposure_time_included__creates_exposure_time_map_using_exposure_time(
        self
    ):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            noise_map_path=test_data_dir + "3_ones.fits",
            primary_beam_path=test_data_dir + "3x3_ones.fits",
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            pixel_scale=0.1,
            exposure_time_map_from_single_value=3.0,
        )

        assert (uv_plane_data.exposure_time_map == 3.0 * np.ones((3,))).all()

    def test__pad_shape_of_primary_beam(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            resized_primary_beam_shape=(9, 9),
            renormalize_primary_beam=False,
        )

        primary_beam_padded_array = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        assert (uv_plane_data.primary_beam == primary_beam_padded_array).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__trim_shape_of_primary_beam(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            resized_primary_beam_shape=(1, 1),
            renormalize_primary_beam=False,
        )

        trimmed_array = np.array([[1.0]])

        assert (uv_plane_data.primary_beam == 5.0 * trimmed_array).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__primary_beam_renormalized_false__does_not_renormalize_primary_beam(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            renormalize_primary_beam=False,
        )

        assert (uv_plane_data.primary_beam == 5.0 * np.ones((3, 3))).all()
        assert (uv_plane_data.noise_map == 3.0 * np.ones((3,))).all()
        assert (uv_plane_data.exposure_time_map == 6.0 * np.ones((3,))).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__primary_beam_renormalized_true__renormalized_primary_beam(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            renormalize_primary_beam=True,
        )

        assert uv_plane_data.primary_beam == pytest.approx(
            (1.0 / 9.0) * np.ones((3, 3)), 1e-2
        )
        assert (uv_plane_data.noise_map == 3.0 * np.ones((3,))).all()
        assert (uv_plane_data.exposure_time_map == 6.0 * np.ones((3,))).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__convert_visibilities_from_electrons_using_exposure_time(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(2, 2),
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            renormalize_primary_beam=False,
            convert_from_electrons=True,
        )

        assert (uv_plane_data.visibilities[:, 0] == np.ones((3,)) / 6.0).all()
        assert (uv_plane_data.visibilities[:, 1] == 2.0 * np.ones((3,)) / 6.0).all()
        assert (uv_plane_data.noise_map == 3.0 * np.ones((3,)) / 6.0).all()
        assert (uv_plane_data.exposure_time_map == 6.0 * np.ones((3,))).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__convert_image_from_adus_using_exposure_time_and_gain(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(2, 2),
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            pixel_scale=0.1,
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            gain=2.0,
            convert_from_adus=True,
        )

        assert (uv_plane_data.visibilities[:, 0] == 2.0 * np.ones((3,)) / 6.0).all()
        assert (
            uv_plane_data.visibilities[:, 1] == 2.0 * 2.0 * np.ones((3,)) / 6.0
        ).all()
        assert (uv_plane_data.noise_map == 2.0 * 3.0 * np.ones((3, 3)) / 6.0).all()
        assert (uv_plane_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert uv_plane_data.pixel_scale == 0.1

    def test__exposure_time_and_exposure_time_map_included__raies_imaging_error(self):

        with pytest.raises(exc.DataException):
            al.load_uv_plane_data_from_fits(
                shape=(7, 7),
                real_visibilities_path=test_data_dir + "3_ones.fits",
                imaginary_visibilities_path=test_data_dir + "3_twos.fits",
                pixel_scale=0.1,
                noise_map_path=test_data_dir + "3x3_threes.fits",
                exposure_time_map_path=test_data_dir + "3x3_ones.fits",
                exposure_time_map_from_single_value=1.0,
            )

    def test__output_all_arrays(self):

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            real_visibilities_path=test_data_dir + "3_ones.fits",
            imaginary_visibilities_path=test_data_dir + "3_twos.fits",
            noise_map_path=test_data_dir + "3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3_sixes.fits",
            u_wavelengths_path=test_data_dir + "3_fours.fits",
            v_wavelengths_path=test_data_dir + "3_fives.fits",
            renormalize_primary_beam=False,
        )

        output_data_dir = "{}/../test_files/array/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        al.output_uv_plane_data_to_fits(
            uv_plane_data=uv_plane_data,
            real_visibilities_path=output_data_dir + "real_visibilities.fits",
            imaginary_visibilities_path=output_data_dir + "imaginary_visibilities.fits",
            noise_map_path=output_data_dir + "noise_map.fits",
            primary_beam_path=output_data_dir + "primary_beam.fits",
            exposure_time_map_path=output_data_dir + "exposure_time_map.fits",
            u_wavelengths_path=output_data_dir + "u_wavelengths.fits",
            v_wavelengths_path=output_data_dir + "v_wavelengths.fits",
            overwrite=True,
        )

        uv_plane_data = al.load_uv_plane_data_from_fits(
            shape=(7, 7),
            pixel_scale=0.1,
            real_visibilities_path=output_data_dir + "real_visibilities.fits",
            imaginary_visibilities_path=output_data_dir + "imaginary_visibilities.fits",
            noise_map_path=output_data_dir + "noise_map.fits",
            primary_beam_path=output_data_dir + "primary_beam.fits",
            exposure_time_map_path=output_data_dir + "exposure_time_map.fits",
            u_wavelengths_path=output_data_dir + "u_wavelengths.fits",
            v_wavelengths_path=output_data_dir + "v_wavelengths.fits",
            renormalize_primary_beam=False,
        )

        assert (uv_plane_data.noise_map == 3.0 * np.ones((3, 3))).all()
        assert (uv_plane_data.primary_beam == 5.0 * np.ones((3, 3))).all()
        assert (uv_plane_data.exposure_time_map == 6.0 * np.ones((3,))).all()
        assert (uv_plane_data.visibilities[:, 0] == np.ones(3)).all()
        assert (uv_plane_data.visibilities[:, 1] == 2.0 * np.ones(3)).all()
        assert (uv_plane_data.noise_map == 3.0 * np.ones(3)).all()
        assert (uv_plane_data.uv_wavelengths[:, 0] == 4.0 * np.ones(3)).all()
        assert (uv_plane_data.uv_wavelengths[:, 1] == 5.0 * np.ones(3)).all()

        assert uv_plane_data.pixel_scale == 0.1



from autolens.pipeline.phase.phase_extensions import HyperGalaxyPhase
from autolens.pipeline.phase.phase import AbstractPhase
from autolens.pipeline.phase.phase_imaging import PhaseImaging
from autolens.pipeline.phase.phase_imaging import PhaseData



import logging

import numpy as np

from autolens import decorator_util
from autolens.array.util import mask_util

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG


@decorator_util.jit()
def sub_mask_1d_index_to_mask_1d_index_from_mask(mask, sub_size):
    """"For pixels on a 2D array of shape (rows, colums), compute a 1D array which, for every unmasked pixel on \
    this 2D array, maps the 1D sub-pixel indexes to their 1D pixel indexes.

    For example, for a sub-grid size of 2, the following mappings from sub-pixels to 2D array pixels are:

    - sub_mask_1d_index_to_mask_1d_index[0] = 0 -> The first sub-pixel maps to the first unmasked pixel on the 2D array.
    - sub_mask_1d_index_to_mask_1d_index[3] = 0 -> The fourth sub-pixel maps to the first unmasked pixel on the 2D array.
    - sub_mask_1d_index_to_mask_1d_index[7] = 1 -> The eighth sub-pixel maps to the second unmasked pixel on the 2D array.

    The term 'grid' is used because the grid is defined as the grid of coordinates on the centre of every \
    pixel on the 2D array. Thus, this array maps sub-pixels on a sub-grid to pixels on a grid.


                     [True, False, True]])
    sub_mask_1d_index_to_mask_1d_index = sub_mask_1d_index_to_mask_1d_index_from_mask(mask=mask, sub_size=2)
    """

    total_sub_pixels = mask_util.total_sub_pixels_from_mask_and_sub_size(
        mask=mask, sub_size=sub_size
    )

    sub_mask_1d_index_to_mask_1d_index = np.zeros(shape=total_sub_pixels)
    mask_1d_index = 0
    sub_mask_1d_index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                for y1 in range(sub_size):
                    for x1 in range(sub_size):
                        sub_mask_1d_index_to_mask_1d_index[
                            sub_mask_1d_index
                        ] = mask_1d_index
                        sub_mask_1d_index += 1

                mask_1d_index += 1

    return sub_mask_1d_index_to_mask_1d_index


def mask_1d_index_to_sub_mask_1d_indexes_from_mask(mask, sub_size):
    """"For pixels on a 2D array of shape (rows, colums), compute a 1D array which, for every unmasked pixel on \
    this 2D array, maps the 1D sub-pixel indexes to their 1D pixel indexes.

    For example, for a sub-grid size of 2, the following mappings from sub-pixels to 2D array pixels are:

    - sub_mask_1d_index_to_mask_1d_index[0] = 0 -> The first sub-pixel maps to the first unmasked pixel on the 2D array.
    - sub_mask_1d_index_to_mask_1d_index[3] = 0 -> The fourth sub-pixel maps to the first unmasked pixel on the 2D array.
    - sub_mask_1d_index_to_mask_1d_index[7] = 1 -> The eighth sub-pixel maps to the second unmasked pixel on the 2D array.

    The term 'grid' is used because the grid is defined as the grid of coordinates on the centre of every \
    pixel on the 2D array. Thus, this array maps sub-pixels on a sub-grid to pixels on a grid.


                     [True, False, True]])
    sub_mask_1d_index_to_mask_1d_index = sub_mask_1d_index_to_mask_1d_index_from_mask(mask=mask, sub_size=2)
    """

    total_pixels = mask_util.total_pixels_from_mask(mask=mask)

    mask_1d_index_to_sub_mask_indexes = [[] for _ in range(total_pixels)]

    sub_mask_1d_index_to_mask_1d_index = sub_mask_1d_index_to_mask_1d_index_from_mask(
        mask=mask, sub_size=sub_size
    ).astype("int")

    for sub_mask_1d_index, mask_1d_index in enumerate(
        sub_mask_1d_index_to_mask_1d_index
    ):
        mask_1d_index_to_sub_mask_indexes[mask_1d_index].append(sub_mask_1d_index)

    return mask_1d_index_to_sub_mask_indexes


@decorator_util.jit()
def submask_index_to_sub_mask_1d_index_from_sub_mask(sub_mask):
    """Create a 2D array which maps every False entry of a 2D mask to its 1D mask array index 2D binned mask. Every \
    True entry is given a value -1.

    This is used as a convenience tool for creating arrays mapping_util between different grids and arrays.

    For example, if we had a 3x4:

    [[False, True, False, False],
     [False, True, False, False],
     [False, False, False, True]]]

    The mask_2d_to_mask_1d array would be:

    [[0, -1, 2, 3],
     [4, -1, 5, 6],
     [7, 8, 9, -1]]

    Parameters
    ----------
    sub_mask : ndarray
        The 2D mask that the mapping_util array is created for.

    Returns
    -------
    ndarray
        The 2D array mapping_util 2D mask entries to their 1D masked array indexes.

    Examples
    --------
    mask_2d = np.full(fill_value=False, shape=(9,9))
    sub_two_to_one = mask_2d_to_mask_1d_index_frommask(mask_2d=mask_2d)
    """

    submask_index_to_1d_index = np.full(fill_value=-1, shape=sub_mask.shape)

    sub_mask_1d_index = 0

    for sub_mask_y in range(sub_mask.shape[0]):
        for sub_mask_x in range(sub_mask.shape[1]):
            if sub_mask[sub_mask_y, sub_mask_x] == False:
                submask_index_to_1d_index[sub_mask_y, sub_mask_x] = sub_mask_1d_index
                sub_mask_1d_index += 1

    return submask_index_to_1d_index


@decorator_util.jit()
def sub_mask_1d_index_to_submask_index_from_mask_and_sub_size(mask, sub_size):
    """Compute a 1D array that maps every unmasked sub-pixel to its corresponding 2d pixel using its (y,x) pixel indexes.

    For example, for a sub-grid size of 2, f pixel [2,5] corresponds to the first pixel in the masked 1D array:

    - The first sub-pixel in this pixel on the 1D array is grid_to_pixel[4] = [2,5]
    - The second sub-pixel in this pixel on the 1D array is grid_to_pixel[5] = [2,6]
    - The third sub-pixel in this pixel on the 1D array is grid_to_pixel[5] = [3,5]

    Parameters
    -----------
    mask : ndarray
        A 2D array of bools, where *False* values are unmasked.
    sub_size : int
        The size of the sub-grid in each mask pixel.

    Returns
    --------
    ndarray
        The 2D blurring mask array whose unmasked values (*False*) correspond to where the mask will have PSF light \
        blurred into them.

    Examples
    --------
    mask = np.array([[True, True, True],
                     [True, False, True]
                     [True, True, True]])

    blurring_mask = blurring_mask_from_mask_and_psf_shape(mask=mask, psf_shape=(3,3))

    """

    total_sub_pixels = mask_util.total_sub_pixels_from_mask_and_sub_size(
        mask=mask, sub_size=sub_size
    )
    sub_mask_1d_index_to_submask_index = np.zeros(shape=(total_sub_pixels, 2))
    sub_mask_1d_index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                for y1 in range(sub_size):
                    for x1 in range(sub_size):
                        sub_mask_1d_index_to_submask_index[sub_mask_1d_index, :] = (
                            (y * sub_size) + y1,
                            (x * sub_size) + x1,
                        )
                        sub_mask_1d_index += 1

    return sub_mask_1d_index_to_submask_index



from astropy import cosmology as cosmo

import autofit as af
from autolens.model.galaxy import galaxy_fit, galaxy_data as gd
from autolens.pipeline.phase.phase import AbstractPhase
from autolens.plotters import visualizer


class PhaseGalaxy(af.AbstractPhase):
    galaxies = af.PhaseProperty("galaxies")

    def __init__(
        self,
        phase_name,
        phase_folders=tuple(),
        galaxies=None,
        use_image=False,
        use_convergence=False,
        use_potential=False,
        use_deflections=False,
        optimizer_class=af.MultiNest,
        sub_size=2,
        pixel_scale_interpolation_grid=None,
        mask_function=None,
        cosmology=cosmo.Planck15,
    ):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit
        models and hyper_galaxies passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_size: int
            The side length of the subgrid
        """

        super(PhaseGalaxy, self).__init__(
            phase_name=phase_name,
            phase_folders=phase_folders,
            optimizer_class=optimizer_class,
        )
        self.cosmology = cosmology
        self.use_image = use_image
        self.use_convergence = use_convergence
        self.use_potential = use_potential
        self.use_deflections = use_deflections
        self.galaxies = galaxies
        self.sub_size = sub_size
        self.pixel_scale_interpolation_grid = pixel_scale_interpolation_grid
        self.mask_function = mask_function

    def run(self, galaxy_data, results=None, mask=None):
        """
        Run this phase.

        Parameters
        ----------
        galaxy_data
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has
            been executed

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """
        analysis = self.make_analysis(
            galaxy_data=galaxy_data, results=results, mask=mask
        )

        self.variable = self.variable.populate(results)
        self.customize_priors(results)
        self.assert_and_save_pickle()

        result = self.run_analysis(analysis)

        return self.make_result(result, analysis)

    def make_analysis(self, galaxy_data, results=None, mask=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying
        functions to allow child classes to change the behaviour of the phase.

        Parameters
        ----------
        galaxy_data
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens: Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a
             set of values
        """

        mask = setup_phase_mask(
            data=galaxy_data[0],
            mask=mask,
            mask_function=self.mask_function,
            inner_mask_radii=None,
        )

        if self.use_image or self.use_convergence or self.use_potential:

            galaxy_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data[0],
                mask=mask,
                pixel_scale_interpolation_grid=self.pixel_scale_interpolation_grid,
                use_image=self.use_image,
                use_convergence=self.use_convergence,
                use_potential=self.use_potential,
                use_deflections_y=self.use_deflections,
                use_deflections_x=self.use_deflections,
            )

            return self.AnalysisSingle(
                galaxy_data=galaxy_data,
                cosmology=self.cosmology,
                image_path=self.optimizer.image_path,
                results=results,
            )

        elif self.use_deflections:

            galaxy_data_y = gd.GalaxyFitData(
                galaxy_data=galaxy_data[0],
                mask=mask,
                pixel_scale_interpolation_grid=self.pixel_scale_interpolation_grid,
                use_image=self.use_image,
                use_convergence=self.use_convergence,
                use_potential=self.use_potential,
                use_deflections_y=self.use_deflections,
                use_deflections_x=False,
            )

            galaxy_data_x = gd.GalaxyFitData(
                galaxy_data=galaxy_data[1],
                mask=mask,
                pixel_scale_interpolation_grid=self.pixel_scale_interpolation_grid,
                use_image=self.use_image,
                use_convergence=self.use_convergence,
                use_potential=self.use_potential,
                use_deflections_y=False,
                use_deflections_x=self.use_deflections,
            )

            return self.AnalysisDeflections(
                galaxy_data_y=galaxy_data_y,
                galaxy_data_x=galaxy_data_x,
                cosmology=self.cosmology,
                image_path=self.optimizer.image_path,
                results=results,
            )

    # noinspection PyAbstractClass
    class Analysis(AbstractPhase.Analysis):
        def __init__(self, cosmology, results, image_path):
            super().__init__(cosmology=cosmology, results=results)
            self.visualizer = visualizer.PhaseGalaxyVisualizer(image_path)

    # noinspection PyAbstractClass
    class AnalysisSingle(Analysis):
        def __init__(self, galaxy_data, cosmology, image_path: str, results=None):
            super().__init__(
                cosmology=cosmology, image_path=image_path, results=results
            )

            self.galaxy_data = galaxy_data

        def fit(self, instance):
            fit = self.fit_for_instance(instance=instance)
            return fit.figure_of_merit

        def visualize(self, instance, during_analysis):
            fit = self.fit_for_instance(instance=instance)

            self.visualizer.plot_galaxy_fit_subplot(fit)

            if during_analysis:
                self.visualizer.plot_fit_individuals(fit)
            else:

                if self.visualizer.plot_ray_tracing_all_at_end_png:
                    self.visualizer.plot_fit_individuals(
                        fit=fit, plot_all=True, image_format="png"
                    )

                if self.visualizer.plot_ray_tracing_all_at_end_fits:
                    self.visualizer.plot_fit_individuals(
                        fit=fit,
                        plot_all=True,
                        image_format="fits",
                        path_suffix="/fits/",
                    )

            return fit

        def fit_for_instance(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in
            this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model
                lens_data itself
            """
            return galaxy_fit.GalaxyFit(
                galaxy_data=self.galaxy_data, model_galaxies=instance.galaxies
            )

    # noinspection PyAbstractClass
    class AnalysisDeflections(Analysis):
        def __init__(
            self, galaxy_data_y, galaxy_data_x, cosmology, image_path, results=None
        ):
            super().__init__(
                cosmology=cosmology, image_path=image_path, results=results
            )

            self.galaxy_data_y = galaxy_data_y
            self.galaxy_data_x = galaxy_data_x

        def fit(self, instance):
            fit_y, fit_x = self.fit_for_instance(instance=instance)
            return fit_y.figure_of_merit + fit_x.figure_of_merit

        def visualize(self, instance, during_analysis):

            fit_y, fit_x = self.fit_for_instance(instance=instance)

            if self.visualizer.plot_galaxy_fit_as_subplot:
                self.visualizer.plot_galaxy_fit_subplot(fit_y, path_suffix="/fit_y_")
                self.visualizer.plot_galaxy_fit_subplot(fit_x, path_suffix="/fit_x_")

            if during_analysis:
                self.visualizer.plot_fit_individuals(fit_y, path_suffix="/fit_y")
                self.visualizer.plot_fit_individuals(fit_x, path_suffix="/fit_x")
            else:
                if self.visualizer.plot_ray_tracing_all_at_end_png:
                    self.visualizer.plot_fit_individuals(
                        fit_y, path_suffix="/fits/fit_y", plot_all=True
                    )
                    self.visualizer.plot_fit_individuals(
                        fit_x, path_suffix="/fits/fit_x", plot_all=True
                    )

                if self.visualizer.plot_ray_tracing_all_at_end_fits:
                    self.visualizer.plot_fit_individuals(
                        fit_y,
                        path_suffix="/fits/fit_y",
                        plot_all=True,
                        image_format="fits",
                    )
                    self.visualizer.plot_fit_individuals(
                        fit_x,
                        path_suffix="/fits/fit_x",
                        plot_all=True,
                        image_format="fits",
                    )

            return fit_y, fit_x

        def fit_for_instance(self, instance):

            fit_y = galaxy_fit.GalaxyFit(
                galaxy_data=self.galaxy_data_y, model_galaxies=instance.galaxies
            )
            fit_x = galaxy_fit.GalaxyFit(
                galaxy_data=self.galaxy_data_x, model_galaxies=instance.galaxies
            )

            return fit_y, fit_x



import numpy as np

import autofit as af
from autolens.model.galaxy import galaxy as g

from autolens.array.mapping import reshape_returned_array


class UVPlaneFit(af.DataFit):
    def __init__(
        self, visibilities, noise_map, mask, model_visibilities, mapping, inversion
    ):

        super().__init__(
            data=visibilities,
            noise_map=noise_map,
            mask=mask,
            model_data=model_visibilities,
        )

        self.mapping = mapping
        self.inversion = inversion

    def visibilities(self):
        return self._data

    def noise_map(self):
        return self._noise_map

    def mask(self):
        return self.mapping.mask

    def signal_to_noise_map(self):
        return self._signal_to_noise_map

    def model_visibilities(self):
        return self._model_data

    def residual_map(self):
        return self._residual_map

    def normalized_residual_map(self):
        return self._normalized_residual_map

    def chi_squared_map(self):
        return self._chi_squared_map

    # @property
    # def likelihood_with_regularization(self):
    #     if self.inversion is not None:
    #         return likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
    #             chi_squared=self.chi_squared,
    #             regularization_term=self.inversion.regularization_term,
    #             noise_normalization=self.noise_normalization,
    #         )
    #
    # @property
    # def evidence(self):
    #     if self.inversion is not None:
    #         return evidence_from_inversion_terms(
    #             chi_squared=self.chi_squared,
    #             regularization_term=self.inversion.regularization_term,
    #             log_curvature_regularization_term=self.inversion.log_det_curvature_reg_matrix_term,
    #             log_regularization_term=self.inversion.log_det_regularization_matrix_term,
    #             noise_normalization=self.noise_normalization,
    #         )

    @property
    def figure_of_merit(self):
        if self.inversion is None:
            return self.likelihood
        # else:
        #     return self.evidence


class LensUVPlaneFit(UVPlaneFit):
    def __init__(
        self,
        tracer,
        grid,
        visibilities,
        noise_map,
        mask,
        inversion,
        model_visibilities,
        transformer,
        positions=None,
    ):

        self.tracer = tracer
        self.grid = grid
        self.transformer = transformer
        self.positions = positions

        super().__init__(
            visibilities=visibilities,
            noise_map=noise_map,
            mask=mask,
            model_visibilities=model_visibilities,
            mapping=grid.mapping,
            inversion=inversion,
        )

    @classmethod
    def from_lens_data_and_tracer(cls, lens_data, tracer):
        """ An  lens fitter, which contains the tracer's used to perform the fit and functions to manipulate \
        the lens data's hyper_galaxies.

        Parameters
        -----------
        tracer : ray_tracing.Tracer
            The tracer, which describes the ray-tracing and strong lens configuration.
        scaled_array_2d_from_array_1d : func
            A function which maps the 1D lens hyper_galaxies to its unmasked 2D array.
        """

        profile_visibilities = tracer.profile_visibilities_from_grid_and_transformer(
            grid=lens_data.grid, transformer=lens_data.transformer
        )

        # profile_subtracted_visibilities_1d = visibilities_1d - blurred_profile_visibilities_1d
        #
        # if not tracer.has_pixelization:

        inversion = None
        model_visibilities = profile_visibilities

        # else:
        #
        #     inversion = tracer.inversion_from_grid_visibilities_1d_noise_map_1d_and_convolver(
        #         grid=lens_uv_plane_data.grid,
        #         visibilities_1d=profile_subtracted_visibilities_1d,
        #         noise_map_1d=noise_map_1d,
        #         convolver=lens_uv_plane_data.convolver,
        #         inversion_uses_border=lens_uv_plane_data.inversion_uses_border,
        #         preload_pixelization_grids_of_planes=lens_uv_plane_data.preload_pixelization_grids_of_planes,
        #     )
        #
        #     model_visibilities_1d = blurred_profile_visibilities_1d + inversion.reconstructed_data_1d

        return cls(
            tracer=tracer,
            visibilities=lens_data.visibilities(),
            noise_map=lens_data.noise_map(return_x2=True),
            mask=lens_data.visibilities_mask,
            model_visibilities=model_visibilities,
            grid=lens_data.grid,
            transformer=lens_data.transformer,
            inversion=inversion,
            positions=lens_data.positions,
        )

    def profile_visibilities(self):
        return self.tracer.tracer.profile_visibilities_from_grid_and_transformer(
            grid=self.grid, transformer=self.transformer, return_in_2d=False
        )

    @reshape_returned_array
    def profile_subtracted_visibilities(self):
        return self.visibilities() - self.profile_visibilities()

    # @property
    # def galaxy_visibilities_1d_dict(self) -> {g.Galaxy: np.ndarray}:
    #     """
    #     A dictionary associating galaxies with their corresponding model visibilities
    #     """
    #     galaxy_visibilities_dict = self.tracer.galaxy_visibilities_dict_from_grid_and_convolver(
    #         grid=self.grid, convolver=self.convolver
    #     )
    #
    #     # TODO : Extend to multiple inversioons across Planes
    #
    #     for plane_index in self.tracer.plane_indexes_with_pixelizations:
    #
    #         galaxy_visibilities_dict.update(
    #             {
    #                 self.tracer.planes[plane_index].galaxies[
    #                     0
    #                 ]: self.inversion.reconstructed_data_1d
    #             }
    #         )
    #
    #     return galaxy_visibilities_dict
    #
    # @property
    # def galaxy_visibilities_2d_dict(self) -> {g.Galaxy: np.ndarray}:
    #     """
    #     A dictionary associating galaxies with their corresponding model visibilities
    #     """
    #
    #     galaxy_visibilities_2d_dict = {}
    #
    #     for galalxy, galaxy_visibilities in self.galaxy_visibilities_1d_dict.items():
    #
    #         galaxy_visibilities_2d_dict[galalxy] = self.grid.mapping.scaled_array_2d_from_array_1d(
    #             array_1d=galaxy_visibilities
    #         )
    #
    #     return galaxy_visibilities_2d_dict

    def model_visibilities_of_planes(self):

        model_visibilities_of_planes = self.tracer.profile_visibilities_of_planes_from_grid_and_transformer(
            grid=self.grid, transformer=self.transformer
        )

        for plane_index in self.tracer.plane_indexes_with_pixelizations:

            model_visibilities_of_planes[
                plane_index
            ] += self.inversion.reconstructed_data_1d

        return model_visibilities_of_planes

    @property
    def total_inversions(self):
        return len(list(filter(None, self.tracer.regularizations_of_planes)))



import time

from autolens.model.profiles import mass_profiles as mp
from autolens.lens import lens_data as ld
from autolens.array import mask as msk

from test.simulation import simulation_util

# Although we could test the deflection angles without using an image (e.g. by just making a grid), we have chosen to
# set this test up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

sub_size = 4
radius_arcsec = 3.0

print("sub grid size = " + str(sub_size))
print("circular mask radius = " + str(radius_arcsec) + "\n")

for data_resolution in ["LSST", "Euclid", "HST", "HST_Up", "AO"]:

    imaging_data = simulation_util.load_test_imaging_data(
        data_type="lens_mass__source_smooth",
        data_resolution=data_resolution,
        psf_shape=(3, 3),
    )
    mask = al.Mask.circular(
        shape=imaging_data.shape,
        pixel_scale=imaging_data.pixel_scale,
        radius_arcsec=radius_arcsec,
    )
    lens_data = al.LensData(imaging_data=imaging_data, mask=mask, sub_size=sub_size)

    print("Deflection angle run times for image type " + data_resolution + "\n")
    print("Number of points = " + str(lens_data.grid.shape[0]) + "\n")

    ### EllipticalIsothermal ###

    mass_profile = al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.0
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalIsothermal time = {}".format(diff))

    ### EllipticalPowerLaw (slope = 1.5) ###

    mass_profile = al.mass_profiles.EllipticalPowerLaw(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.0, slope=1.5
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalPowerLaw (slope = 1.5) time = {}".format(diff))

    ### EllipticalPowerLaw (slope = 2.5) ###

    mass_profile = al.mass_profiles.EllipticalPowerLaw(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.0, slope=2.5
    )

    start = time.time()
    mass_profile.deflections_from_grid(grid=lens_data.grid)
    diff = time.time() - start
    print("EllipticalPowerLaw (slope = 2.5) time = {}".format(diff))



import inspect
import typing
from functools import wraps
from astropy import cosmology as cosmo

import autofit as af
from autolens.model import cosmology_util
from autolens import exc


def convert_units_to_input_units(func):
    """

    Parameters
    ----------
    func : (profiles, *args, **kwargs) -> Object
        A function that requries the units of its input parameters to be checked.

    Returns
    -------
        The original function
    """

    @wraps(func)
    def wrapper(profile, *args, **kwargs):
        """

        Parameters
        ----------
        profile : DimensionsProfile
            The profiles that owns the function

        Returns
        -------
            The DimensionsProfile
        """

        # Extract units of calculation, to convert the input variables and profile to use these units.

        unit_length = (
            kwargs["unit_length"] if "unit_length" in kwargs else profile.unit_length
        )
        unit_luminosity = (
            kwargs["unit_luminosity"]
            if "unit_luminosity" in kwargs
            else profile.unit_luminosity
        )
        unit_mass = kwargs["unit_mass"] if "unit_mass" in kwargs else profile.unit_mass

        # Extract input values which are used for conversions

        cosmology = kwargs["cosmology"] if "cosmology" in kwargs else cosmo.Planck15
        redshift_profile = (
            kwargs["redshift_profile"] if "redshift_profile" in kwargs else None
        )
        redshift_source = (
            kwargs["redshift_source"] if "redshift_source" in kwargs else None
        )
        exposure_time = kwargs["exposure_time"] if "exposure_time" in kwargs else None
        redshift_of_cosmic_average_density = (
            kwargs["redshift_of_cosmic_average_density"]
            if "redshift_of_cosmic_average_density" in kwargs
            else "profile"
        )

        # Use cosmology and redshifts to compute conversion factors.

        if redshift_profile is not None and cosmology is not None:

            kpc_per_arcsec = cosmology_util.kpc_per_arcsec_from_redshift_and_cosmology(
                redshift=redshift_profile, cosmology=cosmology
            )

        else:

            kpc_per_arcsec = None

        if (
            redshift_profile is not None
            and redshift_source is not None
            and cosmology is not None
            and unit_length is not None
            and unit_mass is not None
        ):

            critical_surface_density = cosmology_util.critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
                redshift_0=redshift_profile,
                redshift_1=redshift_source,
                cosmology=cosmology,
                unit_length=unit_length,
                unit_mass=unit_mass,
            )

        else:

            critical_surface_density = None

        if (
            redshift_profile is not None
            and cosmology is not None
            and unit_length is not None
            and unit_mass is not None
            and unit_mass is not "angular"
        ):

            if redshift_of_cosmic_average_density is "profile":
                redshift_calc = redshift_profile
            elif redshift_of_cosmic_average_density is "local":
                redshift_calc = 0.0
            else:
                raise exc.UnitsException(
                    "The redshift of the cosmic average density haas been specified as an invalid "
                    "string. Must be (local | profile)"
                )

            cosmic_average_density = cosmology_util.cosmic_average_density_from_redshift_and_cosmology(
                redshift=redshift_calc,
                cosmology=cosmology,
                unit_length=unit_length,
                unit_mass=unit_mass,
            )

        else:

            cosmic_average_density = None

        # Convert all input parameters to units of input units.

        for key, value in kwargs.items():

            if isinstance(value, Length):
                kwargs[key] = value.convert(
                    unit_length=unit_length, kpc_per_arcsec=kpc_per_arcsec
                )

            if isinstance(value, Luminosity):
                kwargs[key] = value.convert(
                    unit_luminosity=unit_luminosity, exposure_time=exposure_time
                )

            if isinstance(value, Mass):
                kwargs[key] = value.convert(
                    unit_mass=unit_mass,
                    critical_surface_density=critical_surface_density,
                )

            if isinstance(value, MassOverLuminosity):
                kwargs[key] = value.convert(
                    unit_luminosity=unit_luminosity,
                    unit_mass=unit_mass,
                    exposure_time=exposure_time,
                    critical_surface_density=critical_surface_density,
                )

            if isinstance(value, MassOverLength2):
                kwargs[key] = value.convert(
                    unit_length=unit_length,
                    unit_mass=unit_mass,
                    kpc_per_arcsec=kpc_per_arcsec,
                    critical_surface_density=critical_surface_density,
                )

            if isinstance(value, MassOverLength3):

                kwargs[key] = value.convert(
                    unit_length=unit_length,
                    unit_mass=unit_mass,
                    kpc_per_arcsec=kpc_per_arcsec,
                    critical_surface_density=critical_surface_density,
                )

        # Convert profile to input parameter units

        profile = profile.new_profile_with_units_converted(
            unit_length=unit_length,
            unit_luminosity=unit_luminosity,
            unit_mass=unit_mass,
            kpc_per_arcsec=kpc_per_arcsec,
            exposure_time=exposure_time,
            critical_surface_density=critical_surface_density,
        )

        if kpc_per_arcsec is not None:
            kwargs["kpc_per_arcsec"] = kpc_per_arcsec

        if critical_surface_density is not None:
            kwargs["critical_surface_density"] = critical_surface_density

        if cosmic_average_density is not None:
            kwargs["cosmic_average_density"] = cosmic_average_density

        return func(profile, *args, **kwargs)

    return wrapper


class DimensionsProfile(object):
    def __init__(self):

        pass

    def new_profile_with_units_converted(
        self,
        unit_length=None,
        unit_luminosity=None,
        unit_mass=None,
        kpc_per_arcsec=None,
        exposure_time=None,
        critical_surface_density=None,
    ):

        constructor_args = inspect.getfullargspec(self.__init__).args

        def convert(value):
            if unit_length is not None:
                if isinstance(value, Length):
                    return value.convert(unit_length, kpc_per_arcsec)
                if isinstance(value, tuple):
                    return tuple(convert(item) for item in value)
            if unit_luminosity is not None and isinstance(value, Luminosity):
                return value.convert(unit_luminosity, exposure_time)
            if unit_mass is not None and isinstance(value, Mass):
                return value.convert(unit_mass, critical_surface_density)
            if (unit_mass is not None or unit_luminosity is not None) and isinstance(
                value, MassOverLuminosity
            ):
                return value.convert(
                    unit_luminosity, unit_mass, exposure_time, critical_surface_density
                )
            return value

        return self.__class__(
            **{
                key: convert(value)
                for key, value in self.__dict__.items()
                if key in constructor_args
            }
        )

    @property
    def unit_length(self):

        for attr, value in self.__dict__.items():
            if hasattr(value, "unit_length"):
                return value.unit_length

        return None

    @property
    def unit_luminosity(self):

        for attr, value in self.__dict__.items():
            if hasattr(value, "unit_luminosity"):
                return value.unit_luminosity

        return None

    @property
    def unit_mass(self):

        for attr, value in self.__dict__.items():
            if hasattr(value, "unit_mass"):
                return value.unit_mass

        return None


class Length(af.DimensionType):
    def __init__(self, value, unit_length="arcsec"):
        super().__init__(value)
        self.unit_length = unit_length
        self.unit_length_power = 1.0

    @property
    def unit(self):
        return self.unit_length

    def convert(self, unit_length, kpc_per_arcsec=None):
        value = self

        value = convert_length(
            value=value,
            unit_current=self.unit_length,
            unit_new=unit_length,
            power=self.unit_length_power,
            kpc_per_arcsec=kpc_per_arcsec,
        )

        return Length(value=value, unit_length=unit_length)


class Luminosity(af.DimensionType):
    def __init__(self, value, unit_luminosity="eps"):
        super().__init__(value)
        self.unit_luminosity = unit_luminosity
        self.unit_luminosity_power = 1.0

    @property
    def unit(self):
        return self.unit_luminosity

    def convert(self, unit_luminosity, exposure_time=None):
        value = self

        value = convert_luminosity(
            value=value,
            unit_current=self.unit_luminosity,
            unit_new=unit_luminosity,
            power=self.unit_luminosity_power,
            exposure_time=exposure_time,
        )

        return Luminosity(value=value, unit_luminosity=unit_luminosity)


class Mass(af.DimensionType):
    def __init__(self, value, unit_mass="angular"):
        super().__init__(value)
        self.unit_mass = unit_mass
        self.unit_mass_power = 1.0

    @property
    def unit(self):
        return self.unit_mass

    def convert(self, unit_mass, critical_surface_density=None):
        value = self

        value = convert_mass(
            value=value,
            unit_current=self.unit_mass,
            unit_new=unit_mass,
            critical_surface_density=critical_surface_density,
        )

        return Mass(value=value, unit_mass=unit_mass)


class MassOverLuminosity(af.DimensionType):
    def __init__(self, value, unit_luminosity="eps", unit_mass="angular"):

        super().__init__(value)
        self.unit_luminosity = unit_luminosity
        self.unit_luminosity_power = -1.0
        self.unit_mass = unit_mass
        self.unit_mass_power = 1.0

    @property
    def unit(self):
        return self.unit_mass + " / " + self.unit_luminosity

    def convert(
        self,
        unit_luminosity,
        unit_mass,
        exposure_time=None,
        critical_surface_density=None,
    ):

        value = self
        if unit_luminosity is not None:
            value = convert_luminosity(
                value=value,
                unit_current=self.unit_luminosity,
                unit_new=unit_luminosity,
                power=self.unit_luminosity_power,
                exposure_time=exposure_time,
            )
        else:
            unit_luminosity = self.unit_luminosity

        if unit_mass is not None:
            value = convert_mass(
                value=value,
                unit_current=self.unit_mass,
                unit_new=unit_mass,
                critical_surface_density=critical_surface_density,
            )
        else:
            unit_mass = self.unit_mass

        return MassOverLuminosity(
            value=value, unit_mass=unit_mass, unit_luminosity=unit_luminosity
        )


class MassOverLength2(af.DimensionType):
    def __init__(self, value, unit_length="arcsec", unit_mass="angular"):
        super().__init__(value)
        self.unit_length = unit_length
        self.unit_length_power = -2.0
        self.unit_mass = unit_mass
        self.unit_mass_power = 1.0

    @property
    def unit(self):
        return self.unit_mass + " / " + self.unit_length + "^2"

    def convert(
        self, unit_length, unit_mass, kpc_per_arcsec=None, critical_surface_density=None
    ):

        value = self

        if unit_length is not None:
            value = convert_length(
                value=value,
                unit_current=self.unit_length,
                unit_new=unit_length,
                power=self.unit_length_power,
                kpc_per_arcsec=kpc_per_arcsec,
            )
        else:
            unit_length = value.unit_length

        if unit_mass is not None:
            value = convert_mass(
                value=value,
                unit_current=self.unit_mass,
                unit_new=unit_mass,
                critical_surface_density=critical_surface_density,
            )
        else:
            unit_mass = value.unit_mass

        return MassOverLength2(
            value=value, unit_mass=unit_mass, unit_length=unit_length
        )


class MassOverLength3(af.DimensionType):
    def __init__(self, value, unit_length="arcsec", unit_mass="angular"):
        super().__init__(value)
        self.unit_length = unit_length
        self.unit_length_power = -3.0
        self.unit_mass = unit_mass
        self.unit_mass_power = 1.0

    @property
    def unit(self):
        return self.unit_mass + " / " + self.unit_length + "^3"

    def convert(
        self, unit_length, unit_mass, kpc_per_arcsec=None, critical_surface_density=None
    ):

        value = self

        if unit_length is not None:
            value = convert_length(
                value=value,
                unit_current=self.unit_length,
                unit_new=unit_length,
                power=self.unit_length_power,
                kpc_per_arcsec=kpc_per_arcsec,
            )
        else:
            unit_length = value.unit_length

        if unit_mass is not None:
            value = convert_mass(
                value=value,
                unit_current=self.unit_mass,
                unit_new=unit_mass,
                critical_surface_density=critical_surface_density,
            )
        else:
            unit_mass = value.unit_mass

        return MassOverLength3(
            value=value, unit_mass=unit_mass, unit_length=unit_length
        )


Position = typing.Tuple[Length, Length]


def convert_length(value, unit_current, unit_new, power, kpc_per_arcsec):

    if unit_current is not unit_new and kpc_per_arcsec is None:
        raise exc.UnitsException(
            "The length for a value has been requested in new units without a "
            "kpc_per_arcsec conversion factor."
        )

    if unit_current is unit_new:
        return value
    elif unit_current is "arcsec" and unit_new is "kpc":
        return (kpc_per_arcsec ** power) * value
    elif unit_current is "kpc" and unit_new is "arcsec":
        return value / (kpc_per_arcsec ** power)
    else:
        raise exc.UnitsException(
            "The unit specified for the length of a value was an invalid string, you "
            "must use (arcsec | kpc)"
        )


def convert_luminosity(value, unit_current, unit_new, power, exposure_time):
    if unit_current is not unit_new and exposure_time is None:
        raise exc.UnitsException(
            "The luminosity for a value has been requested in new units "
            "without an  exposure time conversion factor."
        )

    if unit_current is unit_new:
        return value
    elif unit_current is "eps" and unit_new is "counts":
        return (exposure_time ** power) * value
    elif unit_current is "counts" and unit_new is "eps":
        return value / (exposure_time ** power)
    else:
        raise exc.UnitsException(
            "The unit specified for the luminosity of a value was an invalid string, you "
            "must use (electrons per second | counts)"
        )


def convert_mass(value, unit_current, unit_new, critical_surface_density):
    if unit_current is not unit_new and critical_surface_density is None:
        raise exc.UnitsException(
            "The mass for a value has been requested in new units "
            "without a critical surface mass density conversion factor."
        )

    if unit_current is unit_new:
        return value
    elif unit_current is "angular" and unit_new is "solMass":
        return critical_surface_density * value
    elif unit_current is "solMass" and unit_new is "angular":
        return value / critical_surface_density
    else:
        raise exc.UnitsException(
            "The unit specified for the mass of a value was an invalid string, you "
            "must use (angular | solMass)"
        )



from autolens import decorator_util
import numpy as np

from autolens.array.util import array_util, mask_util
from autolens.array.mapping_util import mask_mapping_util


@decorator_util.jit()
def padded_binning_shape_2d_from_shape_2d_and_bin_up_factor(shape_2d, bin_up_factor):

    shape_remainder = (shape_2d[0] % bin_up_factor, shape_2d[1] % bin_up_factor)

    if shape_remainder[0] != 0 and shape_remainder[1] != 0:
        shape_pad = (
            bin_up_factor - shape_remainder[0],
            bin_up_factor - shape_remainder[1],
        )
    elif shape_remainder[0] != 0 and shape_remainder[1] == 0:
        shape_pad = (bin_up_factor - shape_remainder[0], 0)
    elif shape_remainder[0] == 0 and shape_remainder[1] != 0:
        shape_pad = (0, bin_up_factor - shape_remainder[1])
    else:
        shape_pad = (0, 0)

    return (shape_2d[0] + shape_pad[0], shape_2d[1] + shape_pad[1])


@decorator_util.jit()
def padded_binning_array_2d_from_array_2d_and_bin_up_factor(
    array_2d, bin_up_factor, pad_value=0.0
):
    """If an array is to be binned up, but the dimensions are not divisible by the bin-up factor, this routine pads \
    the array to make it divisible.

    For example, if the array is shape (5,5) and the bin_up_factor is 2, this routine will pad the array to shape \
    (6,6).

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is padded.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).
    pad_value : float
        If the array is padded, the value the padded edge values are filled in using.

    Returns
    -------
    ndarray
        The 2D array that is padded before binning up.

    Examples
    --------
    array_2d = np.ones((5,5))
    padded_array_2d = padded_array_2d_for_binning_up_with_bin_up_factor( \
        array_2d=array_2d, bin_up_factor=2, pad_value=0.0)
    """

    padded_binning_shape_2d = padded_binning_shape_2d_from_shape_2d_and_bin_up_factor(
        shape_2d=array_2d.shape, bin_up_factor=bin_up_factor
    )

    return array_util.resized_array_2d_from_array_2d_and_resized_shape(
        array_2d=array_2d, resized_shape=padded_binning_shape_2d, pad_value=pad_value
    )


@decorator_util.jit()
def binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
    array_2d, bin_up_factor
):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their mean value to determine \
     the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the mean of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), the array is first \
    padded to make the division work. One must be careful of edge effects in this case.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is binned up.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The binned up 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = bin_up_array_2d_using_mean(array_2d=array_2d, bin_up_factor=2)
    """

    padded_binning_array_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=array_2d, bin_up_factor=bin_up_factor
    )

    binned_array_2d = np.zeros(
        shape=(
            padded_binning_array_2d.shape[0] // bin_up_factor,
            padded_binning_array_2d.shape[1] // bin_up_factor,
        )
    )

    for y in range(binned_array_2d.shape[0]):
        for x in range(binned_array_2d.shape[1]):
            value = 0.0
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y * bin_up_factor + y1
                    padded_x = x * bin_up_factor + x1
                    value += padded_binning_array_2d[padded_y, padded_x]

            binned_array_2d[y, x] = value / (bin_up_factor ** 2.0)

    return binned_array_2d


@decorator_util.jit()
def binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
    array_2d, bin_up_factor
):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their quadrature value to \
    determine the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the quadrature of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), the array is first \
    padded to make the division work. One must be careful of edge effects in this case.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is binned up.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The binned up 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = bin_up_array_2d_using_quadrature(array_2d=array_2d, bin_up_factor=2)
    """

    padded_binning_array_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=array_2d, bin_up_factor=bin_up_factor
    )

    binned_array_2d = np.zeros(
        shape=(
            padded_binning_array_2d.shape[0] // bin_up_factor,
            padded_binning_array_2d.shape[1] // bin_up_factor,
        )
    )

    for y in range(binned_array_2d.shape[0]):
        for x in range(binned_array_2d.shape[1]):
            value = 0.0
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y * bin_up_factor + y1
                    padded_x = x * bin_up_factor + x1
                    value += padded_binning_array_2d[padded_y, padded_x] ** 2.0

            binned_array_2d[y, x] = np.sqrt(value) / (bin_up_factor ** 2.0)

    return binned_array_2d


@decorator_util.jit()
def binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(array_2d, bin_up_factor):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their sum value to determine \
     the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the sum of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), the array is first \
    padded to make the division work. One must be careful of edge effects in this case.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is binned up.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The binned up 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = bin_up_array_2d_using_sum(array_2d=array_2d, bin_up_factor=2)
    """

    padded_binning_array_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=array_2d, bin_up_factor=bin_up_factor
    )

    binned_array_2d = np.zeros(
        shape=(
            padded_binning_array_2d.shape[0] // bin_up_factor,
            padded_binning_array_2d.shape[1] // bin_up_factor,
        )
    )

    for y in range(binned_array_2d.shape[0]):
        for x in range(binned_array_2d.shape[1]):
            value = 0.0
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y * bin_up_factor + y1
                    padded_x = x * bin_up_factor + x1
                    value += padded_binning_array_2d[padded_y, padded_x]

            binned_array_2d[y, x] = value

    return binned_array_2d


@decorator_util.jit()
def binned_upmask_frommask_and_bin_up_factor(mask_2d, bin_up_factor):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their sum value to determine \
     the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the sum of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), an array is first \
    extracted around the centre of that array.


    Parameters
    ----------
    mask_2d : ndarray
        The 2D array that is resized.
    new_shape : (int, int)
        The (y,x) new pixel dimension of the trimmed array.
    origin : (int, int)
        The oigin of the resized array, e.g. the central pixel around which the array is extracted.

    Returns
    -------
    ndarray
        The resized 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = resize_array_2d(array_2d=array_2d, new_shape=(2,2), origin=(2, 2))
    """

    paddedmask = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True
    )

    binnedmask = np.zeros(
        shape=(
            paddedmask.shape[0] // bin_up_factor,
            paddedmask.shape[1] // bin_up_factor,
        )
    )

    for y in range(binnedmask.shape[0]):
        for x in range(binnedmask.shape[1]):
            value = True
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y * bin_up_factor + y1
                    padded_x = x * bin_up_factor + x1
                    if paddedmask[padded_y, padded_x] == False:
                        value = False

            binnedmask[y, x] = value

    return binnedmask


@decorator_util.jit()
def paddedmask_to_mask_1d_index_frommask_and_bin_up_factor(mask_2d, bin_up_factor):
    """Create a 2D array which maps every False entry of a 2D mask to its 1D mask array index 2D binned mask. Every \
    True entry is given a value -1.

    This uses the function *paddedmask_to_mask_1d_index*, see this method for a more detailed description of the \
    mapping_util.

    This function first pads the mask using the same padding when computed a binned up mask.

    Parameters
    ----------
    mask_2d : ndarray
        The 2D mask that the mapping_util array is created for.

    Returns
    -------
    ndarray
        The 2D array mapping_util padded 2D mask entries to their 1D masked array indexes.

    Examples
    --------
    mask_2d = np.full(fill_value=False, shape=(9,9))
    mask_2d_to_mask_1d_index = paddedmask_to_mask_1d_index_frommask(mask_2d=mask_2d)
    """

    paddedmask = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True
    )

    return mask_mapping_util.submask_index_to_sub_mask_1d_index_from_sub_mask(
        sub_mask=paddedmask
    )


@decorator_util.jit()
def paddedmask_to_binned_mask_1d_index_frommask_and_bin_up_factor(
    mask_2d, bin_up_factor
):
    """Create a 2D array which maps every False entry of a 2D mask to its 1D binned mask index (created using the \
    *binned_upmask_frommask_and_bin_up_factor* method).

    We create an array the same shape as the 2D mask (after padding for the binnning up procedure), where each entry \
    gives the binned up mask's 1D masked array index.
    
    This is used as a convenience tool for creating arrays mapping_util between different grids and arrays.
    
    For example, if we had a 4x4 mask:
    
    [[False, False, False, False],
     [False, False, False, False],
     [ True,  True, False, False],
     [ True,  True, False, False]]
     
    For a bin_up_factor of 2, the resulting binned up mask is as follows (noting there is no padding in this example):
    
    [[False, False],
      [True, False]

    The mask_2d_to_binned_mask_1d_index is therefore:

    [[ 0,  0, 1, 1],
     [ 0,  0, 1, 1],
     [-1, -1, 2, 2],
     [-1, -1, 2, 2]]

    Parameters
    ----------
    mask_2d : ndarray
        The 2D mask that the binned mask 1d indexes are computing using
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The 2D array mapping_util 2D mask entries to their 1D binned masked array indexes.

    Examples
    --------
    mask_2d = np.full(fill_value=False, shape=(9,9))
    mask_to_binnedmask =
        mask_2d_to_binned_mask_1d_index_frommask_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
    """

    paddedmask = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True
    )

    binnedmask = binned_upmask_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    paddedmask_index_to_binned_mask_1d_index = np.full(
        fill_value=-1, shape=paddedmask.shape
    )

    binned_mask_1d_index = 0

    for bin_y in range(binnedmask.shape[0]):
        for bin_x in range(binnedmask.shape[1]):
            if binnedmask[bin_y, bin_x] == False:
                for bin_y1 in range(bin_up_factor):
                    for bin_x1 in range(bin_up_factor):
                        mask_y = bin_y * bin_up_factor + bin_y1
                        mask_x = bin_x * bin_up_factor + bin_x1
                        if paddedmask[mask_y, mask_x] == False:
                            paddedmask_index_to_binned_mask_1d_index[
                                mask_y, mask_x
                            ] = binned_mask_1d_index

                binned_mask_1d_index += 1

    return paddedmask_index_to_binned_mask_1d_index


@decorator_util.jit()
def masked_array_1d_to_binned_masked_array_1d_frommask_and_bin_up_factor(
    mask_2d, bin_up_factor
):
    """Create a 1D array which maps every (padded) masked index to its corresponding 1D index in the binned 1D \
    mask.

    This uses the convenience tools *paddedmask_to_mask_1d* and *paddedmask_to_binned_mask_1d* to \
    make the calculation simpler.

    For example, if we had a 4x4 mask:

    [[False, False, False, False],
     [False, False, False, False],
     [ True,  True, False, False],
     [ True,  True, False, False]]

    For a bin_up_factor of 2, the resulting binned up mask is as follows (noting there is no padding in this example):

    [[False, False],
      [True, False]

    The mask_2d_to_mask_1d_index is therefore:

    [[ 0,  1, 2,  3],
     [ 4,  5, 6,  7],
     [-1, -1, 8,  9],
     [-1, -1, 10, 11]]

    And the mask_2d_to_binned_mask_1d_index is therefore:

    [[ 0,  0, 1, 1],
     [ 0,  0, 1, 1],
     [-1, -1, 2, 2],
     [-1, -1, 2, 2]]

    Therefore, the binned_masked_array_1d_to_masked_array_1d would be:

    [0, 0, 1, 1, 0, 0, 1, 1, 2, 2, 2, 2]

        This tells us that:
     - The first mask pixel maps to the first binned masked pixel (e.g. the 1D index of mask_2d after binning up).
     - The second mask pixel maps to the first binned masked pixel (e.g. the 1D index of mask_2d after binning up)
     - The third mask pixel maps to the second masked pixel (e.g. the 1D index of mask_2d after binning up)

    Parameters
    ----------
    mask_2d : ndarray
        The 2D mask that the binned mask 1d index mappings are computed using
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The 1D array mapping_util 1D binned mask entries to their corresponding 1D masked array index.

    Examples
    --------
    mask_2d = np.full(fill_value=False, shape=(9,9))
    mask_to_binnedmask =
        binned_masked_array_1d_to_masked_array_1d_frommask_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
    """

    paddedmask = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True
    )

    total_masked_pixels = mask_util.total_pixels_from_mask(mask=paddedmask)

    masked_array_1d_to_binned_masked_array_1d = np.zeros(shape=total_masked_pixels)

    paddedmask_to_mask_1d_index = paddedmask_to_mask_1d_index_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    paddedmask_to_binned_mask_1d_index = paddedmask_to_binned_mask_1d_index_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    for mask_y in range(paddedmask_to_mask_1d_index.shape[0]):
        for mask_x in range(paddedmask_to_mask_1d_index.shape[1]):
            if paddedmask_to_mask_1d_index[mask_y, mask_x] >= 0:
                padded_mask_index = paddedmask_to_mask_1d_index[mask_y, mask_x]
                binned_mask_1d_index = paddedmask_to_binned_mask_1d_index[
                    mask_y, mask_x
                ]
                masked_array_1d_to_binned_masked_array_1d[
                    padded_mask_index
                ] = binned_mask_1d_index

    return masked_array_1d_to_binned_masked_array_1d


@decorator_util.jit()
def binned_masked_array_1d_to_masked_array_1d_frommask_and_bin_up_factor(
    mask_2d, bin_up_factor
):
    """Create a 1D array which maps every (padded) binned masked index to its correspond 1D index in the original 2D \
    mask that was binned up.

    Array indexing starts from the top-left and goes rightwards and downwards. The top-left pixel of each mask is \
    used before binning up.

    This uses the convenience tools *paddedmask_to_mask_1d* and *paddedmask_to_binned_mask_1d* to \
    make the calculation simpler.

    For example, if we had a 4x4 mask:

    [[False, False, False, False],
     [False, False, False, False],
     [ True,  True, False, False],
     [ True,  True, False, False]]

    For a bin_up_factor of 2, the resulting binned up mask is as follows (noting there is no padding in this example):

    [[False, False],
      [True, False]

    The mask_2d_to_mask_1d_index is therefore:

    [[ 0,  1, 2,  3],
     [ 4,  5, 6,  7],
     [-1, -1, 8,  9],
     [-1, -1, 10, 11]]

    And the mask_2d_to_binned_mask_1d_index is therefore:

    [[ 0,  0, 1, 1],
     [ 0,  0, 1, 1],
     [-1, -1, 2, 2],
     [-1, -1, 2, 2]]

    Therefore, the binned_masked_array_1d_to_masked_array_1d would be:

    [0, 2, 8]

    This tells us that:
     - The first binned mask pixel maps to the first masked pixel (e.g. the 1D index of mask_2d).
     - The second binned mask pixel maps to the third masked pixel (e.g. the 1D index of mask_2d)
     - The third binned mask pixel maps to the ninth masked pixel (e.g. the 1D index of mask_2d)

    Parameters
    ----------
    mask_2d : ndarray
        The 2D mask that the binned mask 1d index mappings are computed using
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The 1D array mapping_util 1D binned mask entries to their corresponding 1D masked array index.

    Examples
    --------
    mask_2d = np.full(fill_value=False, shape=(9,9))
    mask_to_binnedmask =
        binned_masked_array_1d_to_masked_array_1d_frommask_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
    """

    binned_upmask = binned_upmask_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    total_binned_masked_pixels = mask_util.total_pixels_from_mask(mask=binned_upmask)

    binned_masked_array_1d_to_masked_array_1d = -1 * np.ones(total_binned_masked_pixels)

    paddedmask_to_mask_1d_index = paddedmask_to_mask_1d_index_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    paddedmask_to_binned_mask_1d_index = paddedmask_to_binned_mask_1d_index_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    for mask_y in range(paddedmask_to_mask_1d_index.shape[0]):
        for mask_x in range(paddedmask_to_mask_1d_index.shape[1]):
            if paddedmask_to_mask_1d_index[mask_y, mask_x] >= 0:
                binned_mask_index = paddedmask_to_binned_mask_1d_index[mask_y, mask_x]
                if binned_masked_array_1d_to_masked_array_1d[binned_mask_index] == -1:
                    padded_mask_index = paddedmask_to_mask_1d_index[mask_y, mask_x]
                    binned_masked_array_1d_to_masked_array_1d[
                        binned_mask_index
                    ] = padded_mask_index

    return binned_masked_array_1d_to_masked_array_1d


@decorator_util.jit()
def binned_masked_array_1d_to_masked_array_1d_all_frommask_and_bin_up_factor(
    mask_2d, bin_up_factor
):
    """Create a 2D array which maps every (padded) binned masked index to all of the corresponding 1D indexes of the \
    the original 2D mask that was binned up.

    Array indexing starts from the top-left and goes rightwards and downwards. The top-left pixel of each mask is \
    used before binning up. Minus one's are used for mapping_util which go to masked values with True.

    This uses the convenience tools *paddedmask_to_mask_1d* and *paddedmask_to_binned_mask_1d* to \
    make the calculation simpler.

    For example, if we had a 4x4 mask:

    [[False, False, False, False],
     [False, False, False, False],
     [ True,  True, False, False],
     [ True,  True, True, False]]

    For a bin_up_factor of 2, the resulting binned up mask is as follows (noting there is no padding in this example):

    [[False, False],
      [True, False]

    The mask_2d_to_mask_1d_index is therefore:

    [[ 0,  1, 2,  3],
     [ 4,  5, 6,  7],
     [-1, -1, 8,  9],
     [-1, -1, -1, 10]]

    And the mask_2d_to_binned_mask_1d_index is therefore:

    [[ 0,  0, 1, 1],
     [ 0,  0, 1, 1],
     [-1, -1, 2, 2],
     [-1, -1, 2, 2]]

    Therefore, the binned_masked_array_1d_to_masked_array_1d_all would be:

    [[0, 1, 4, 5],
     [2, 3, 6, 7]]
     [8, 9, 10, -1]]

    This tells us that:
     - The first binned mask pixel maps to the first, second, fifth and sixth masked pixels.
     - The second binned mask pixel maps to the third, fourth, seventh and eighth masked pixels
     - The third binned mask pixel maps to the ninth, tenth and eleventh masked pixels (The fourth masked pixel it \
       maps to is a *True* value and therefore masked.)

    Parameters
    ----------
    mask_2d : ndarray
        The 2D mask that the binned mask 1d index mappings are computed using
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The 1D array mapping_util 1D binned mask entries to their corresponding 1D masked array index.

    Examples
    --------
    mask_2d = np.full(fill_value=False, shape=(9,9))
    mask_to_binnedmask =
        binned_masked_array_1d_to_masked_array_1d_frommask_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
    """

    binned_upmask = binned_upmask_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    total_binned_masked_pixels = mask_util.total_pixels_from_mask(mask=binned_upmask)

    binned_masked_array_1d_to_masked_array_1d_all = -1 * np.ones(
        (total_binned_masked_pixels, bin_up_factor ** 2)
    )

    binned_masked_array_1d_sizes = np.zeros(total_binned_masked_pixels)

    paddedmask_to_mask_1d_index = paddedmask_to_mask_1d_index_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    paddedmask_to_binned_mask_1d_index = paddedmask_to_binned_mask_1d_index_frommask_and_bin_up_factor(
        mask_2d=mask_2d, bin_up_factor=bin_up_factor
    )

    for mask_y in range(paddedmask_to_mask_1d_index.shape[0]):
        for mask_x in range(paddedmask_to_mask_1d_index.shape[1]):
            if paddedmask_to_mask_1d_index[mask_y, mask_x] >= 0:

                binned_mask_index = paddedmask_to_binned_mask_1d_index[mask_y, mask_x]
                binned_mask_count = int(binned_masked_array_1d_sizes[binned_mask_index])
                padded_mask_index = paddedmask_to_mask_1d_index[mask_y, mask_x]

                binned_masked_array_1d_to_masked_array_1d_all[
                    binned_mask_index, binned_mask_count
                ] = padded_mask_index

                binned_masked_array_1d_sizes[binned_mask_index] += 1

    return binned_masked_array_1d_to_masked_array_1d_all, binned_masked_array_1d_sizes



import autofit as af
import autolens as al
from test.integration.tests import runner

test_type = "lens__source_inversion"
test_name = "lens_mass__source_adaptive_brightness__hyper"
data_type = "lens_mass__source_smooth"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class Phase1(al.PhaseImaging):
        def customize_priors(self, results):
            self.galaxies.source.light.sersic_index = af.UniformPrior(3.9, 4.1)

    phase1 = Phase1(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 50
    phase1.optimizer.sampling_efficiency = 0.8

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):
            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.galaxies.source.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.source.hyper_galaxy
            )

    phase2 = InversionPhase(
        phase_name="phase_2_weighted_regularization",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=phase1.result.constant.galaxies.lens.mass,
                shear=phase1.result.constant.galaxies.lens.shear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pixelizations.VoronoiBrightnessImage,
                regularization=al.regularization.AdaptiveBrightness,
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 30
    phase2.optimizer.sampling_efficiency = 0.8

    phase2 = phase2.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=True,
    )

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):
            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.galaxies.source.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.source.hyper_galaxy
            )

    phase3 = InversionPhase(
        phase_name="phase_3",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=phase1.result.variable.galaxies.lens.mass,
                shear=phase1.result.variable.galaxies.lens.shear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=phase2.result.constant.galaxies.source.pixelization,
                regularization=phase2.result.constant.galaxies.source.regularization,
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 40
    phase3.optimizer.sampling_efficiency = 0.8

    phase3 = phase3.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=True,
    )

    return al.PipelineImaging(name, phase1, phase2, phase3)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import numpy as np

import autofit as af
from autolens.model.galaxy import galaxy as g

from autolens.array.mapping import reshape_returned_array


class ImagingFit(af.DataFit):
    def __init__(self, image, noise_map, mask, model_image, mapping, inversion):

        super().__init__(
            data=image, noise_map=noise_map, mask=mask, model_data=model_image
        )

        self.mapping = mapping
        self.inversion = inversion

    @reshape_returned_array
    def image(self, return_in_2d=True, return_masked=True):
        return self._data

    @reshape_returned_array
    def noise_map(self, return_in_2d=True, return_masked=True):
        return self._noise_map

    @property
    def mask(self):
        return self.mapping.mask

    @reshape_returned_array
    def signal_to_noise_map(self, return_in_2d=True, return_masked=True):
        return self._signal_to_noise_map

    @reshape_returned_array
    def model_image(self, return_in_2d=True):
        return self._model_data

    @reshape_returned_array
    def residual_map(self, return_in_2d=True):
        return self._residual_map

    @reshape_returned_array
    def normalized_residual_map(self, return_in_2d=True):
        return self._normalized_residual_map

    @reshape_returned_array
    def chi_squared_map(self, return_in_2d=True):
        return self._chi_squared_map

    @property
    def likelihood_with_regularization(self):
        if self.inversion is not None:
            return likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=self.chi_squared,
                regularization_term=self.inversion.regularization_term,
                noise_normalization=self.noise_normalization,
            )

    @property
    def evidence(self):
        if self.inversion is not None:
            return evidence_from_inversion_terms(
                chi_squared=self.chi_squared,
                regularization_term=self.inversion.regularization_term,
                log_curvature_regularization_term=self.inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=self.inversion.log_det_regularization_matrix_term,
                noise_normalization=self.noise_normalization,
            )

    @property
    def figure_of_merit(self):
        if self.inversion is None:
            return self.likelihood
        else:
            return self.evidence


class LensImagingFit(ImagingFit):
    def __init__(
        self,
        tracer,
        grid,
        blurring_grid,
        image,
        noise_map,
        mask,
        inversion,
        model_image,
        convolver,
        positions=None,
    ):

        self.tracer = tracer
        self.grid = grid
        self.blurring_grid = blurring_grid
        self.psf = convolver.psf
        self.convolver = convolver
        self.positions = positions

        super().__init__(
            image=image,
            noise_map=noise_map,
            mask=mask,
            model_image=model_image,
            mapping=grid.mapping,
            inversion=inversion,
        )

    @classmethod
    def from_lens_data_and_tracer(
        cls, lens_data, tracer, hyper_image_sky=None, hyper_background_noise=None
    ):
        """ An  lens fitter, which contains the tracer's used to perform the fit and functions to manipulate \
        the lens data's hyper_galaxies.

        Parameters
        -----------
        tracer : ray_tracing.Tracer
            The tracer, which describes the ray-tracing and strong lens configuration.
        scaled_array_2d_from_array_1d : func
            A function which maps the 1D lens hyper_galaxies to its unmasked 2D array.
        """

        image_1d = image_1d_from_lens_data_and_hyper_image_sky(
            lens_data=lens_data, hyper_image_sky=hyper_image_sky
        )

        noise_map_1d = noise_map_1d_from_lens_data_tracer_and_hyper_backkground_noise(
            lens_data=lens_data,
            tracer=tracer,
            hyper_background_noise=hyper_background_noise,
        )

        blurred_profile_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
            grid=lens_data.grid,
            convolver=lens_data.convolver,
            blurring_grid=lens_data.blurring_grid,
            return_in_2d=False,
        )

        profile_subtracted_image_1d = image_1d - blurred_profile_image_1d

        if not tracer.has_pixelization:

            inversion = None
            model_image_1d = blurred_profile_image_1d

        else:

            inversion = tracer.inversion_from_grid_image_1d_noise_map_1d_and_convolver(
                grid=lens_data.grid,
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=noise_map_1d,
                convolver=lens_data.convolver,
                inversion_uses_border=lens_data.inversion_uses_border,
                preload_pixelization_grids_of_planes=lens_data.preload_pixelization_grids_of_planes,
            )

            model_image_1d = blurred_profile_image_1d + inversion.reconstructed_data_1d

        return cls(
            tracer=tracer,
            image=image_1d,
            noise_map=noise_map_1d,
            mask=lens_data._mask_1d,
            model_image=model_image_1d,
            grid=lens_data.grid,
            blurring_grid=lens_data.blurring_grid,
            convolver=lens_data.convolver,
            inversion=inversion,
            positions=lens_data.positions,
        )

    @reshape_returned_array
    def blurred_profile_image(self, return_in_2d=True):
        return self.tracer.blurred_profile_image_from_grid_and_psf(
            grid=self.grid,
            psf=self.psf,
            blurring_grid=self.blurring_grid,
            return_in_2d=False,
        )

    @reshape_returned_array
    def profile_subtracted_image(self, return_in_2d=True):
        return self.image(return_in_2d=False) - self.blurred_profile_image(
            return_in_2d=False
        )

    @property
    def galaxy_model_image_1d_dict(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """
        galaxy_model_image_dict = self.tracer.galaxy_blurred_profile_image_dict_from_grid_and_convolver(
            grid=self.grid,
            convolver=self.convolver,
            blurring_grid=self.blurring_grid,
            return_in_2d=False,
        )

        # TODO : Extend to multiple inversioons across Planes

        for plane_index in self.tracer.plane_indexes_with_pixelizations:

            galaxy_model_image_dict.update(
                {
                    self.tracer.planes[plane_index].galaxies[
                        0
                    ]: self.inversion.reconstructed_data_1d
                }
            )

        return galaxy_model_image_dict

    @property
    def galaxy_model_image_2d_dict(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_model_image_2d_dict = {}

        for galalxy, galaxy_image in self.galaxy_model_image_1d_dict.items():

            galaxy_model_image_2d_dict[
                galalxy
            ] = self.grid.mapping.scaled_array_2d_from_array_1d(array_1d=galaxy_image)

        return galaxy_model_image_2d_dict

    def model_images_of_planes(self, return_in_2d=True):

        model_images_of_planes = self.tracer.blurred_profile_images_of_planes_from_grid_and_psf(
            grid=self.grid,
            psf=self.psf,
            blurring_grid=self.blurring_grid,
            return_in_2d=return_in_2d,
        )

        for plane_index in self.tracer.plane_indexes_with_pixelizations:

            if return_in_2d:
                model_images_of_planes[
                    plane_index
                ] += self.inversion.reconstructed_data_2d
            else:
                model_images_of_planes[
                    plane_index
                ] += self.inversion.reconstructed_data_1d

        return model_images_of_planes

    @property
    def total_inversions(self):
        return len(list(filter(None, self.tracer.regularizations_of_planes)))

    @property
    def unmasked_blurred_profile_image(self):
        return self.tracer.unmasked_blurred_profile_image_from_grid_and_psf(
            grid=self.grid, psf=self.psf
        )

    @property
    def unmasked_blurred_profile_image_of_planes(self):
        return self.tracer.unmasked_blurred_profile_image_of_planes_from_grid_and_psf(
            grid=self.grid, psf=self.psf
        )

    @property
    def unmasked_blurred_profile_image_of_planes_and_galaxies(self):
        return self.tracer.unmasked_blurred_profile_image_of_planes_and_galaxies_from_grid_and_psf(
            grid=self.grid, psf=self.psf
        )


def image_1d_from_lens_data_and_hyper_image_sky(lens_data, hyper_image_sky):

    if hyper_image_sky is not None:
        return hyper_image_sky.image_scaled_sky_from_image(image=lens_data._image_1d)
    else:
        return lens_data._image_1d


def noise_map_1d_from_lens_data_tracer_and_hyper_backkground_noise(
    lens_data, tracer, hyper_background_noise
):

    if hyper_background_noise is not None:
        noise_map_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
            noise_map=lens_data._noise_map_1d
        )
    else:
        noise_map_1d = lens_data._noise_map_1d

    hyper_noise_map_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
        noise_map_1d=lens_data._noise_map_1d
    )

    if hyper_noise_map_1d is not None:
        noise_map_1d = noise_map_1d + hyper_noise_map_1d
        if lens_data.hyper_noise_map_max is not None:
            noise_map_1d[
                noise_map_1d > lens_data.hyper_noise_map_max
            ] = lens_data.hyper_noise_map_max

    return noise_map_1d


def likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
    chi_squared, regularization_term, noise_normalization
):
    """Compute the likelihood of an inversion's fit to the datas, including a regularization term which \
    comes from an inversion:

    Likelihood = -0.5*[Chi_Squared_Term + Regularization_Term + Noise_Term] (see functions above for these definitions)

    Parameters
    ----------
    chi_squared : float
        The chi-squared term of the inversion's fit to the observed datas.
    regularization_term : float
        The regularization term of the inversion, which is the sum of the difference between reconstructed \
        flux of every pixel multiplied by the regularization coefficient.
    noise_normalization : float
        The normalization noise_map-term for the observed datas's noise-map.
    """
    return -0.5 * (chi_squared + regularization_term + noise_normalization)


def evidence_from_inversion_terms(
    chi_squared,
    regularization_term,
    log_curvature_regularization_term,
    log_regularization_term,
    noise_normalization,
):
    """Compute the evidence of an inversion's fit to the datas, where the evidence includes a number of \
    terms which quantify the complexity of an inversion's reconstruction (see the *inversion* module):

    Likelihood = -0.5*[Chi_Squared_Term + Regularization_Term + Log(Covariance_Regularization_Term) -
                       Log(Regularization_Matrix_Term) + Noise_Term]

    Parameters
    ----------
    chi_squared : float
        The chi-squared term of the inversion's fit to the observed datas.
    regularization_term : float
        The regularization term of the inversion, which is the sum of the difference between reconstructed \
        flux of every pixel multiplied by the regularization coefficient.
    log_curvature_regularization_term : float
        The log of the determinant of the sum of the curvature and regularization matrices.
    log_regularization_term : float
        The log of the determinant o the regularization matrix.
    noise_normalization : float
        The normalization noise_map-term for the observed datas's noise-map.
    """
    return -0.5 * (
        chi_squared
        + regularization_term
        + log_curvature_regularization_term
        - log_regularization_term
        + noise_normalization
    )



import autolens as al
import os

import pytest


@pytest.fixture(name="uv_plane_plotter_path")
def make_uv_plane_plotter_setup():
    uv_plane_plotter_path = "{}/../../test_files/plotting/uv_plane/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    return uv_plane_plotter_path


def test__individual_attributes_are_output(
    uv_plane_data_7, uv_plane_plotter_path, plot_patch
):
    al.uv_plane_plotters.plot_visibilities(
        uv_plane_data=uv_plane_data_7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=uv_plane_plotter_path,
        output_format="png",
    )

    assert uv_plane_plotter_path + "uv_plane_visibilities.png" in plot_patch.paths

    al.uv_plane_plotters.plot_u_wavelengths(
        uv_plane_data=uv_plane_data_7,
        output_path=uv_plane_plotter_path,
        output_format="png",
    )

    assert uv_plane_plotter_path + "uv_plane_u_wavelengths.png" in plot_patch.paths

    al.uv_plane_plotters.plot_v_wavelengths(
        uv_plane_data=uv_plane_data_7,
        output_path=uv_plane_plotter_path,
        output_format="png",
    )

    assert uv_plane_plotter_path + "uv_plane_v_wavelengths.png" in plot_patch.paths

    al.uv_plane_plotters.plot_primary_beam(
        uv_plane_data=uv_plane_data_7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=uv_plane_plotter_path,
        output_format="png",
    )

    assert uv_plane_plotter_path + "uv_plane_primary_beam.png" in plot_patch.paths

    al.uv_plane_plotters.plot_uv_plane_subplot(
        uv_plane_data=uv_plane_data_7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=uv_plane_plotter_path,
        output_format="png",
    )

    assert uv_plane_plotter_path + "uv_plane_data.png" in plot_patch.paths


def test__uv_plane_individuals__output_dependent_on_input(
    uv_plane_data_7, general_config, uv_plane_plotter_path, plot_patch
):
    al.uv_plane_plotters.plot_uv_plane_individual(
        uv_plane_data=uv_plane_data_7,
        should_plot_visibilities=True,
        should_plot_u_wavelengths=False,
        should_plot_v_wavelengths=True,
        should_plot_primary_beam=True,
        output_path=uv_plane_plotter_path,
        output_format="png",
    )

    assert uv_plane_plotter_path + "uv_plane_visibilities.png" in plot_patch.paths

    assert not uv_plane_plotter_path + "uv_plane_u_wavelengths.png" in plot_patch.paths

    assert uv_plane_plotter_path + "uv_plane_v_wavelengths.png" in plot_patch.paths

    assert uv_plane_plotter_path + "uv_plane_primary_beam.png" in plot_patch.paths



# from autolens.array import grids, fourier_transform as ft
# from autolens.data import uv_plane
#
# import numpy as np
# import time
# import os
#
# test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))
# data_path = test_path + "data/uvplane/"
#
# uv_plane_data = uv_plane.load_uv_plane_data_from_fits(
#     image_path=data_path + "image.fits",
#     pixel_scale=0.05,
#     psf_path=data_path + "psf.fits",
#     noise_map_path=data_path + "noise_map.fits",
#     primary_beam_path=data_path + "primary_beam.fits",
#     real_visibilities_path=data_path + "real_visibilities.fits",
#     imaginary_visibilities_path=data_path + "imaginary_visibilities.fits",
#     noise_map_path=data_path + "visibilities_noise_map.fits",
#     u_wavelengths_path=data_path + "u_wavelengths.fits",
#     v_wavelengths_path=data_path + "v_wavelengths.fits",
#     renormalize_psf=False,
#     renormalize_primary_beam=False,
# )
#
# print(uv_plane_data.uv_wavelengths)
# stop
#
# repeats = 1
#
# visibilities = len(uv_plane_data.visibilities)
#
# grid = grids.Grid.from_shape_pixel_scale_and_sub_size(
#     shape=uv_plane_data.image.shape,
#     pixel_scale=uv_plane_data.pixel_scale,
#     sub_size=1,
# )
# grid_radians = grid.in_radians
#
# image_pixels = grid.shape[0] * grid.shape[1]
# source_pixels = 1000
#
# shape_data = 8 * visibilities
# shape_preloads = visibilities * image_pixels * 2
# shape_mapping_matrix = visibilities * source_pixels
#
# total_shape = shape_data + shape_preloads + shape_mapping_matrix
#
# print("Data Memory Use (GB) = " + str(shape_data * 8e-9))
# print("PreLoad Memory Use (GB) = " + str(shape_preloads * 8e-9))
# print("Mapping Matrix Memory Use (GB) = " + str(shape_mapping_matrix * 8e-9))
# print("Total Memory Use (GB) = " + str(total_shape * 8e-9))
# print()
#
# image_1d = np.ones(shape=(image_pixels))
#
# transformer = ft.Transformer(
#     uv_wavelengths=uv_plane_data.uv_wavelengths, grid_radians=grid_radians
# )
#
# print(transformer.preload_real_transforms[0, :])
# print(transformer.preload_real_transforms[1, :])
# print(transformer.preload_real_transforms[2, :])
# stop
#
#
# start = time.time()
# for i in range(repeats):
#     transformer.real_visibilities_from_image_1d(image_1d=image_1d)
# diff = time.time() - start
# print("Real Visibilities Time = {}".format(diff / repeats))
#
# start = time.time()
# for i in range(repeats):
#     transformer.real_visibilities_via_preload_from_image(image_1d=image_1d)
# diff = time.time() - start
# print("Real Visibilities PreLoad Time = {}".format(diff / repeats))



import autolens as al
import os

import pytest


@pytest.fixture(name="phase_plotter_path")
def make_phase_plotter_setup():
    return "{}/../../test_files/plotting/phase/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


def test__plot_imaging_for_phase(
    imaging_data_7x7, mask_7x7, general_config, phase_plotter_path, plot_patch
):
    al.phase_plotters.plot_imaging_for_phase(
        imaging_data=imaging_data_7x7,
        mask=mask_7x7,
        positions=None,
        units="arcsec",
        zoom_around_mask=True,
        extract_array_from_mask=True,
        should_plot_as_subplot=True,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_psf=True,
        should_plot_signal_to_noise_map=False,
        should_plot_absolute_signal_to_noise_map=False,
        should_plot_potential_chi_squared_map=True,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "imaging_data.png" in plot_patch.paths
    assert phase_plotter_path + "imaging/imaging_image.png" in plot_patch.paths
    assert phase_plotter_path + "imaging/imaging_noise_map.png" not in plot_patch.paths
    assert phase_plotter_path + "imaging/imaging_psf.png" in plot_patch.paths
    assert (
        phase_plotter_path + "imaging/imaging_signal_to_noise_map.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "imaging/imaging_absolute_signal_to_noise_map.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "imaging/imaging_potential_chi_squared_map.png"
        in plot_patch.paths
    )


def test__plot_ray_tracing_for_phase__dependent_on_input(
    tracer_x2_plane_7x7, sub_grid_7x7, mask_7x7, phase_plotter_path, plot_patch
):
    al.phase_plotters.plot_ray_tracing_for_phase(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        during_analysis=False,
        mask=mask_7x7,
        positions=None,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        units="arcsec",
        should_plot_as_subplot=True,
        should_plot_all_at_end_png=False,
        should_plot_all_at_end_fits=False,
        should_plot_image=True,
        should_plot_source_plane=True,
        should_plot_convergence=False,
        should_plot_potential=True,
        should_plot_deflections=False,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "tracer.png" in plot_patch.paths
    assert (
        phase_plotter_path + "ray_tracing/tracer_profile_image.png" in plot_patch.paths
    )
    assert (
        phase_plotter_path + "ray_tracing/tracer_source_plane.png" in plot_patch.paths
    )
    assert (
        phase_plotter_path + "ray_tracing/tracer_convergence.png"
        not in plot_patch.paths
    )
    assert phase_plotter_path + "ray_tracing/tracer_potential.png" in plot_patch.paths
    assert (
        phase_plotter_path + "ray_tracing/tracer_deflections_y.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "ray_tracing/tracer_deflections_x.png"
        not in plot_patch.paths
    )


def test__lens_fit_for_phase__source_and_lens__depedent_on_input(
    lens_imaging_fit_x2_plane_7x7, phase_plotter_path, plot_patch
):
    al.phase_plotters.plot_lens_imaging_fit_for_phase(
        fit=lens_imaging_fit_x2_plane_7x7,
        during_analysis=False,
        should_plot_mask=True,
        positions=None,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        units="arcsec",
        should_plot_image_plane_pix=True,
        should_plot_all_at_end_png=False,
        should_plot_all_at_end_fits=False,
        should_plot_fit_as_subplot=True,
        should_plot_fit_of_planes_as_subplot=False,
        should_plot_inversion_as_subplot=False,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_residual_map=False,
        should_plot_normalized_residual_map=True,
        should_plot_chi_squared_map=True,
        should_plot_pixelization_residual_map=False,
        should_plot_pixelization_normalized_residual_map=False,
        should_plot_pixelization_chi_squared_map=False,
        should_plot_pixelization_regularization_weights=False,
        should_plot_subtracted_images_of_planes=True,
        should_plot_model_images_of_planes=False,
        should_plot_plane_images_of_planes=True,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "lens_fit.png" in plot_patch.paths
    assert phase_plotter_path + "lens_fit/fit_image.png" in plot_patch.paths
    assert phase_plotter_path + "lens_fit/fit_noise_map.png" not in plot_patch.paths
    assert (
        phase_plotter_path + "lens_fit/fit_signal_to_noise_map.png"
        not in plot_patch.paths
    )
    assert phase_plotter_path + "lens_fit/fit_model_image.png" in plot_patch.paths
    assert phase_plotter_path + "lens_fit/fit_residual_map.png" not in plot_patch.paths
    assert (
        phase_plotter_path + "lens_fit/fit_normalized_residual_map.png"
        in plot_patch.paths
    )
    assert phase_plotter_path + "lens_fit/fit_chi_squared_map.png" in plot_patch.paths
    assert (
        phase_plotter_path + "lens_fit/fit_subtracted_image_of_plane_0.png"
        in plot_patch.paths
    )
    assert (
        phase_plotter_path + "lens_fit/fit_subtracted_image_of_plane_1.png"
        in plot_patch.paths
    )
    assert (
        phase_plotter_path + "lens_fit/fit_model_image_of_plane_0.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "lens_fit/fit_model_image_of_plane_1.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "lens_fit/fit_plane_image_of_plane_0.png"
        in plot_patch.paths
    )
    assert (
        phase_plotter_path + "lens_fit/fit_plane_image_of_plane_1.png"
        in plot_patch.paths
    )


def test__hyper_images_for_phase__source_and_lens__depedent_on_input(
    hyper_model_image_7x7, binned_grid_7x7, mask_7x7, phase_plotter_path, plot_patch
):
    al.phase_plotters.plot_hyper_images_for_phase(
        hyper_model_image_2d=hyper_model_image_7x7,
        hyper_galaxy_image_2d_path_dict=None,
        binned_hyper_galaxy_image_2d_path_dict=None,
        mask=mask_7x7,
        binned_grid=binned_grid_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        units="arcsec",
        should_plot_hyper_model_image=True,
        should_plot_hyper_galaxy_images=False,
        should_plot_binned_hyper_galaxy_images=False,
        visualize_path=phase_plotter_path,
    )

    assert (
        phase_plotter_path + "hyper_galaxies/hyper_model_image.png" in plot_patch.paths
    )
    assert (
        phase_plotter_path + "hyper_galaxies/hyper_galaxy_images.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "hyper_galaxies/hyper_galaxy_cluster_images.png"
        not in plot_patch.paths
    )
